from .base import Base


class Alerting(Base):
    def __init__(self, client):
        super(Alerting, self).__init__(client)
        self.client = client

    def get_alertrule(self, folder_name, alertrule_name):
        """
        :param folder_name:
        :param alertrule_name:
        :return:
        """
        get_alertrule_path = "/ruler/grafana/api/v1/rules/%s/%s" % (folder_name, alertrule_name)
        return self.client.GET(get_alertrule_path)

    def get_managedalerts_all(self, datasource="grafanacloud-prom"):
        """ """
        get_managedalerts_path = "/prometheus/%s/api/v1/rules" % datasource
        return self.client.GET(get_managedalerts_path)

    def create_alertrule(self, folder_name, alertrule):
        """
        :param folder_name:
        :param alertrule:
        :return:
        """
        create_alertrule_path = "/ruler/grafana/api/v1/rules/%s" % folder_name
        return self.client.POST(create_alertrule_path, json=alertrule)

    def update_alertrule(self, folder_name, alertrule):
        """
        @param folder_name:
        @param alertrule:
        @return:
        """

        update_alertrule_path = "/ruler/grafana/api/v1/rules/%s" % folder_name
        return self.client.POST(update_alertrule_path, json=alertrule)

    def delete_alertrule(self, folder_name, alertrule_name):
        """
        :param folder_name:
        :param alertrule_name:
        @return:
        """

        delete_alertrule_path = "/ruler/grafana/api/v1/rules/%s/%s" % (folder_name, alertrule_name)
        return self.client.DELETE(delete_alertrule_path)



"""
https://grafana.com/docs/grafana/latest/developers/http_api/serviceaccount/
https://grafana.com/docs/grafana/latest/administration/service-accounts/
https://grafana.com/docs/grafana/latest/developers/http_api/create-api-tokens-for-org/
"""

from .base import Base


class ServiceAccount(Base):
    def __init__(self, client):
        super(ServiceAccount, self).__init__(client)
        self.client = client
        self.path = "/serviceaccounts"

    def get(self, service_account_id):
        """
        Get service account by id.
        https://grafana.com/docs/grafana/latest/developers/http_api/serviceaccount/#get-a-service-account-by-id

        :param service_account_id:
        :return:
        """
        get_actual_user_path = "/serviceaccounts/%s?accesscontrol=true" % (service_account_id)
        return self.client.GET(get_actual_user_path)

    def create(self, service_account):
        """
        Create service account.
        https://grafana.com/docs/grafana/latest/developers/http_api/serviceaccount/#create-service-account

        :param service_account: {"name": "string", "role": "string"}
        :return:
        """
        create_service_account_path = "/serviceaccounts/"
        return self.client.POST(create_service_account_path, json=service_account)

    def update(self, service_account_id, service_account):
        """
        Update service account by id.
        https://grafana.com/docs/grafana/latest/developers/http_api/serviceaccount/#update-service-account

        :param service_account_id:
        :param service_account: {"name": "string", "role": "string"}
        :return:
        """
        path = "/serviceaccounts/%s" % service_account_id
        return self.client.PATCH(path, json=service_account)

    def delete(self, service_account_id):
        """
        Delete service account.
        https://grafana.com/docs/grafana/latest/developers/http_api/serviceaccount/#delete-service-account

        :param service_account_id:
        :return:
        """

        delete_service_account_path = "/serviceaccounts/%s" % (service_account_id)
        return self.client.DELETE(delete_service_account_path)

    def get_tokens(self, service_account_id):
        """
        Get service account tokens.
        https://grafana.com/docs/grafana/latest/developers/http_api/serviceaccount/#get-service-account-tokens

        :param service_account_id:
        :return:
        """
        service_account_tokens_path = "/serviceaccounts/%s/tokens" % (service_account_id)
        return self.client.GET(service_account_tokens_path)

    def create_token(self, service_account_id, content):
        """
        Create service account tokens
        https://grafana.com/docs/grafana/latest/developers/http_api/serviceaccount/#create-service-account-tokens

        :param service_account_id:
        :param service_account_name:
        :return:
        """
        create_service_account_token_path = "/serviceaccounts/%s/tokens" % (service_account_id)
        return self.client.POST(create_service_account_token_path, json=content)

    def delete_token(self, service_account_id, service_account_token_id):
        """
        Delete service account tokens.
        https://grafana.com/docs/grafana/latest/developers/http_api/serviceaccount/#delete-service-account-tokens

        :param service_account_id:
        :param service_account_token_id:
        :return:
        """
        delete_service_account_token_path = "/serviceaccounts/%s/tokens/%s" % (
            service_account_id,
            service_account_token_id,
        )
        return self.client.DELETE(delete_service_account_token_path)

    def search(self, query=None, page=None, perpage=None):
        """

        :return:
        """
        list_of_sa = []
        sa_on_page = None
        show_sa_path = "/serviceaccounts/search"
        params = []
        if query:
            params.append("query=%s" % query)

        if page:
            iterate = False
            params.append("page=%s" % page)
        else:
            iterate = True
            params.append("page=%s")
            page = 1

        if perpage:
            params.append("perpage=%s" % perpage)

        show_sa_path += "?"
        show_sa_path += "&".join(params)
        if iterate:
            while True:
                url = show_sa_path % page
                sa_on_page = self.client.GET(url)
                list_of_sa.append(sa_on_page)
                if not sa_on_page["serviceAccounts"]:
                    break
                page += 1
        else:
            sa_on_page = self.client.GET(show_sa_path)
            list_of_sa.append(sa_on_page)

        return list_of_sa

    def search_one(self, service_account_name=""):
        """
        Find a single service account by name. Raises errors on multiple or no matches.

        :param service_account_name:
        :return:
        """
        s = self.search(query=service_account_name)[0]
        if s["totalCount"] == 1:
            return s["serviceAccounts"][0]
        elif s["totalCount"] > 1:
            raise ValueError("More than one service account matched")
        else:
            raise ValueError("No service account matched")



from .base import Base


class DashboardVersions(Base):
    """
    About
    =====
    Access Grafana "Dashboard Versions" API.

    Reference
    =========
    https://grafana.com/docs/grafana/latest/developers/http_api/dashboard_versions/
    """

    def __init__(self, client):
        super(DashboardVersions, self).__init__(client)
        self.client = client

    @staticmethod
    def api_path(dashboard_id: int = None, dashboard_uid: str = None):
        if dashboard_id is not None:
            path = f"/dashboards/id/{dashboard_id}"
        elif dashboard_uid is not None:
            path = f"/dashboards/uid/{dashboard_uid}"
        else:
            raise LookupError("Either dashboard_id or dashboard_uid is required")
        return path

    def get_dashboard_versions(
        self, dashboard_id: int = None, dashboard_uid: str = None, limit: int = None, start: int = None
    ):
        api_path = self.api_path(dashboard_id=dashboard_id, dashboard_uid=dashboard_uid)
        dashboard_versions_path = f"{api_path}/versions"

        query_args = {}
        if limit is not None:
            query_args["limit"] = limit
        if start is not None:
            query_args["start"] = start

        return self.client.GET(dashboard_versions_path, data=query_args)

    def get_dashboard_versions_by_id(self, dashboard_id: int = None, limit: int = None, start: int = None):
        return self.get_dashboard_versions(dashboard_id=dashboard_id, limit=limit, start=start)

    def get_dashboard_versions_by_uid(self, dashboard_uid: str = None, limit: int = None, start: int = None):
        return self.get_dashboard_versions(dashboard_uid=dashboard_uid, limit=limit, start=start)

    def get_dashboard_version(self, dashboard_id: int = None, dashboard_uid: str = None, version_id: int = None):
        api_path = self.api_path(dashboard_id=dashboard_id, dashboard_uid=dashboard_uid)
        dashboard_version_path = f"{api_path}/versions/{version_id}"

        if version_id is None:
            raise LookupError("version_id is required")

        return self.client.GET(dashboard_version_path)

    def get_dashboard_version_by_id(self, dashboard_id: int = None, version_id: int = None):
        return self.get_dashboard_version(dashboard_id=dashboard_id, version_id=version_id)

    def get_dashboard_version_by_uid(self, dashboard_uid: int = None, version_id: int = None):
        return self.get_dashboard_version(dashboard_uid=dashboard_uid, version_id=version_id)

    def restore_dashboard(self, dashboard_id: int = None, dashboard_uid: str = None, version_id: int = None):
        api_path = self.api_path(dashboard_id=dashboard_id, dashboard_uid=dashboard_uid)
        restore_dashboard_path = f"{api_path}/restore"

        if version_id is None:
            raise LookupError("version_id is required")

        return self.client.POST(restore_dashboard_path, json={"version": version_id})

    def restore_dashboard_by_id(self, dashboard_id: int = None, version_id: int = None):
        return self.restore_dashboard(dashboard_id=dashboard_id, version_id=version_id)

    def restore_dashboard_by_uid(self, dashboard_uid: str = None, version_id: int = None):
        return self.restore_dashboard(dashboard_uid=dashboard_uid, version_id=version_id)

    def calculate_diff(
        self,
        base_dashboard_id: int,
        base_version_id: int,
        new_dashboard_id: int,
        new_version_id: int,
        diff_type: str = "json",
    ):
        """
        Compares two dashboard versions by calculating the JSON diff of them.

        :param diff_type: the type of diff to return. Can be "json" or "basic".
        """
        calculate_diff_path = "/dashboards/calculate-diff"

        if diff_type not in ["json", "basic"]:
            raise LookupError("diff_type must be either 'json' or 'basic'")

        return self.client.POST(
            calculate_diff_path,
            json={
                "base": {"dashboardId": base_dashboard_id, "version": base_version_id},
                "new": {"dashboardId": new_dashboard_id, "version": new_version_id},
                "diffType": diff_type,
            },
        )



from sys import modules

import niquests
import urllib3

modules["requests"] = niquests
modules["requests.adapters"] = niquests.adapters
modules["requests.exceptions"] = niquests.exceptions
modules["requests.packages.urllib3"] = urllib3

import requests_mock  # noqa: E402


class Mocker(requests_mock.Mocker):
    """
    This monkeypatch class aim to fix a bug in "requests_mock"
    The library does not add "application/json" header in the response mock,
    when it's clearly about a json response. Niquests does not like it and require
    (as per RFC) a proper content type header to proceed.
    """

    def request(self, *args, **kwargs):
        if "headers" not in kwargs:
            kwargs["headers"] = {}
        if "json" in kwargs and kwargs["json"] is not None:
            kwargs["headers"]["Content-Type"] = "application/json"
        return self.register_uri(*args, **kwargs)


setattr(requests_mock, "Mocker", Mocker)

__all__ = ("requests_mock",)



from copy import deepcopy

ELASTICSEARCH_DATASOURCE = {
    "id": 44,
    "uid": "34inf2sdc",
    "name": "Elasticsearch",
    "type": "elasticsearch",
    "access": "proxy",
    "database": "bazqux",
    "url": "http://localhost:9200",
    "jsonData": {
        "esVersion": "7.10.0",
        "timeField": "@timestamp",
    },
}

GRAPHITE_DATASOURCE = {
    "id": 48,
    "uid": "wr47rz34e",
    "name": "Graphite",
    "type": "graphite",
    "access": "proxy",
}

INFLUXDB1_DATASOURCE = {
    "id": 43,
    "uid": "9ac78sdc2",
    "name": "InfluxDB",
    "type": "influxdb",
    "access": "proxy",
    "url": "http://localhost:8086",
    "jsonData": {"httpMode": "POST", "version": "InfluxQL"},
}

JAEGER_DATASOURCE = {
    "id": 53,
    "uid": "DbtFe237k",
    "name": "Jaeger",
    "type": "jaeger",
    "access": "proxy",
}

LOKI_DATASOURCE = {
    "id": 54,
    "uid": "vCyglaq7z",
    "name": "Loki",
    "type": "loki",
    "access": "proxy",
}

MSSQL_DATASOURCE = {
    "id": 56,
    "uid": "0pueH83nz",
    "name": "MSSQL",
    "type": "mssql",
    "access": "proxy",
}

MYSQL_DATASOURCE = {
    "id": 51,
    "uid": "7CpzLp37z",
    "name": "MariaDB",
    "type": "mysql",
    "access": "proxy",
    "url": "localhost:3306",
}

OPENTSDB_DATASOURCE = {
    "id": 51,
    "uid": "hkuk5h3nk",
    "name": "OpenTSDB",
    "type": "opentsdb",
    "access": "proxy",
}

POSTGRES_DATASOURCE = {
    "id": 50,
    "uid": "v2KYBt37k",
    "name": "PostgreSQL",
    "type": "postgres",
    "access": "proxy",
    "url": "localhost:5432",
    "jsonData": {
        "postgresVersion": 1200,
        "sslmode": "disable",
    },
}

PROMETHEUS_DATASOURCE = {
    "id": 42,
    "uid": "h8KkCLt7z",
    "orgId": 1,
    "name": "Prometheus",
    "type": "prometheus",
    "typeName": "Prometheus",
    "typeLogoUrl": "public/app/plugins/datasource/prometheus/img/prometheus_logo.svg",
    "access": "proxy",
    "url": "http://localhost:9090",
    "password": "",
    "user": "",
    "database": "",
    "basicAuth": False,
    "isDefault": True,
    "jsonData": {"httpMethod": "POST"},
    "readOnly": False,
}

SIMPLE_JSON_DATASOURCE = {
    "id": 47,
    "uid": "rw783ds3e",
    "name": "SimpleJson",
    "type": "grafana-simple-json-datasource",
    "access": "proxy",
}

SIMPOD_JSON_DATASOURCE = {
    "id": 49,
    "uid": "oie238af3",
    "name": "SimpodJson",
    "type": "simpod-json-datasource",
    "access": "proxy",
}

SUNANDMOON_DATASOURCE = {
    "id": 46,
    "uid": "239fasva4",
    "name": "SunAndMoon",
    "type": "fetzerch-sunandmoon-datasource",
    "access": "proxy",
    "jsonData": {
        "latitude": 42.42,
        "longitude": 84.84,
    },
}
SUNANDMOON_DATASOURCE_INCOMPLETE = deepcopy(SUNANDMOON_DATASOURCE)
del SUNANDMOON_DATASOURCE_INCOMPLETE["jsonData"]["latitude"]

TEMPO_DATASOURCE = {
    "id": 55,
    "uid": "aTk86s3nk",
    "name": "Tempo",
    "type": "tempo",
    "access": "proxy",
}

TESTDATA_DATASOURCE = {
    "id": 45,
    "uid": "439fngqr2",
    "name": "Testdata",
    "type": "testdata",
    "access": "proxy",
}

ZIPKIN_DATASOURCE = {
    "id": 57,
    "uid": "3sXIv8q7k",
    "name": "Zipkin",
    "type": "zipkin",
    "access": "proxy",
}

PERMISSION_DATASOURCE = {
    "datasourceId": 42,
    "enabled": True,
    "permissions": [
        {
            "id": 42,
            "datasourceId": 42,
            "teamId": 32,
            "teamAvatarUrl": "/avatar/c5e5ac09a2af1dec9a828d002c6bd4ea",
            "team": "a12",
            "permission": 1,
            "permissionName": "Query",
            "isManaged": True,
            "created": "2022-11-30T15:57:58Z",
            "updated": "2022-11-30T15:57:58Z",
        }
    ],
}

PROMETHEUS_DATA_RESPONSE = {
    "status": "success",
    "data": {
        "resultType": "matrix",
        "result": [
            {
                "metric": {"__name__": "up", "instance": "localhost:9090", "job": "prometheus"},
                "values": [
                    [1644164339, "1"],
                    [1644164399, "1"],
                    [1644164459, "1"],
                    [1644164519, "1"],
                    [1644164579, "1"],
                    [1644164639, "1"],
                ],
            }
        ],
    },
}


DATAFRAME_RESPONSE_EMPTY = {"results": {"test": {"frames": []}}}
DATAFRAME_RESPONSE_INVALID = {"results": {"test": {"frames": "foobar"}}}
DATAFRAME_RESPONSE_HEALTH_SELECT1 = {
    "results": {"test": {"frames": [{"schema": {"meta": {"executedQueryString": "SELECT 1"}}}]}}
}

# TODO: Reflect a real Grafana response for Elasticsearch here.
DATAFRAME_RESPONSE_HEALTH_ELASTICSEARCH_VALID = {"bazqux": {}}

DATAFRAME_RESPONSE_HEALTH_PROMETHEUS = {
    "results": {
        "test": {
            "frames": [
                {
                    "schema": {
                        "name": "1+1",
                        "refId": "test",
                        "meta": {
                            "type": "timeseries-many",
                            "custom": {"resultType": "matrix"},
                            "executedQueryString": "Expr: 1+1\nStep: 15s",
                        },
                        "fields": [
                            {
                                "name": "Time",
                                "type": "time",
                                "typeInfo": {"frame": "time.Time"},
                                "config": {"interval": 15000},
                            },
                            {
                                "name": "Value",
                                "type": "number",
                                "typeInfo": {"frame": "float64", "nullable": True},
                                "labels": {},
                                "config": {"displayNameFromDS": "1+1"},
                            },
                        ],
                    },
                    "data": {"values": [[0], [2]]},
                }
            ]
        }
    }
}



import sys
import unittest

if sys.version_info < (3, 7, 0):
    raise unittest.SkipTest("Imports will fail on Python 3.6. Skipping all tests in test_package_version.py")

from grafana_client.version import __version__, __version_tuple__


class PackageTestCase(unittest.TestCase):
    def test_package_version(self):
        self.assertIsInstance(__version__, str)
        self.assertGreaterEqual(__version_tuple__, (0, 0, 1))



"""
About
=====

Example program for submitting a query to a data source.

While the ``--query`` option is obligatory, the ``--attr`` option is not.
It will be used to designate the ``database`` name, with e.g. InfluxDB,
or the ``queryType`` name, with e.g. Loki.

Synopsis
========
::
    # Query the InfluxDB 1.x data source on `play.grafana.org`.
    python examples/datasource-smartquery.py --uid=000000002 --query="SHOW RETENTION POLICIES on _internal"
    python examples/datasource-smartquery.py --uid=000000002 --attr "database:site" --query="SHOW FIELD KEYS"
    python examples/datasource-smartquery.py --uid 000000002 --query "SELECT sum(\"value\") FROM \"logins.count\" WHERE time >= now() - 5m and time <= now() GROUP BY time(10s) fill(null)"

    # Query the InfluxDB+flux on https://play.grafana.org
    # fix me : query invalid !
    python examples/datasource-smartquery.py \
        --uid M3k6ZPrnz \
        --query "from(bucket: \"HyperEncabulator\")\\n  |> range(start: v.timeRangeStart, stop: v.timeRangeStop)\\n    |> filter(fn: (r) => r[\"_measurement\"] == \"TemperatureData\")\\n    |> filter(fn: (r) => r[\"MeasType\"] == \"setpoint\" or r[\"MeasType\"] == \"actual\")\\n    |> filter(fn: (r) => r[\"Tank\"] == \"A5\")\\n    |> filter(fn: (r) => r[\"_field\"] == \"Temperature\")\\n    |> aggregateWindow(every: v.windowPeriod, fn: mean, createEmpty: false)\\n    |> yield(name: \"mean\")"
    # fix-me
        --query "from(bucket: \"example-bucket\")\\n  |> range(start: -1h)\\n  |> filter(fn: (r) => r._measurement == \"example-measurement\" and r._field == \"example-field\")"

    # Query the Graphite datasource on https://play.grafana.org
    python examples/datasource-smartquery.py --uid 000000001 --query "aliasByNode(apps.fakesite.web_server_02.counters.requests.count,2)"

    # Query the Prometheus data source on `play.grafana.org`.
    python examples/datasource-smartquery.py \
        --uid=000000008 \
        --query='time() - node_boot_time_seconds{job="node", instance=~"demo.do.prometheus.io:.*"}'

    # Query the Loki data source on `play.grafana.org`.
    python examples/datasource-smartquery.py \
        --uid=NX9d1VH7k \
        --query "bytes_over_time({filename=~\".+/json_access.+\", job=~\".*\", instance=~\".*\"}[5m])" \
        --attr "queryType:instant"
    python examples/datasource-smartquery.py \
        --uid=NX9d1VH7k \
        --query "sum by (status) (count_over_time({filename=~\".+/json_access.+\", job=~\".*\", instance=~\".*\"} | json |  __error__=\"\" [5m]))" \
        --attr "legendFormat:HTTP Status {{status}}"

"""  # noqa: E501

import json
import logging
from optparse import OptionParser

import niquests

from grafana_client import GrafanaApi
from grafana_client.model import DatasourceIdentifier
from grafana_client.util import setup_logging

logger = logging.getLogger(__name__)


def run(grafana: GrafanaApi):
    parser = OptionParser()
    parser.add_option("--uid", dest="uid", help="Data source UID")
    # replace by --attr "database:<store>" or --attr "metric:<metric>"
    # parser.add_option("--store", dest="store", help="Database or metric to be addressed")
    parser.add_option("--query", dest="query", help="Query expression")
    parser.add_option("--attr", dest="attrs", action="append", help="key:value pair attribute to send into query")
    (options, args) = parser.parse_args()
    if not options.uid:
        parser.error("Option --uid required")
    if not options.query:
        parser.error("Option --query required")

    attributes = {}
    if options.attrs is not None and len(options.attrs) > 0:
        for attr in options.attrs:
            (key, val) = attr.split(":")
            attributes[key] = val

    # replace double backslash chars by their values \\n => \n
    query = bytes(options.query, "utf-8").decode("unicode_escape")
    # Query the data source by UID.
    response = grafana.datasource.smartquery(
        DatasourceIdentifier(uid=options.uid),
        query,
        attrs=attributes,
    )

    # Display the outcome in JSON format.
    print(json.dumps(response, indent=2))


if __name__ == "__main__":
    setup_logging(level=logging.DEBUG)

    # Connect to Grafana instance and run health check.
    grafana_client = GrafanaApi.from_url("https://play.grafana.org/")

    try:
        grafana_client.connect()
    except niquests.exceptions.ConnectionError:
        logger.exception("Connecting to Grafana failed")
        raise SystemExit(1)

    run(grafana_client)



import niquests
import niquests.auth
from niquests import HTTPError, Timeout
from niquests.exceptions import JSONDecodeError

DEFAULT_TIMEOUT: float = 5.0
DEFAULT_SESSION_POOL_SIZE: int = 10


class GrafanaException(Exception):
    def __init__(self, status_code, response, message):
        self.status_code = status_code
        self.response = response
        self.message = message
        # Backwards compatible with implementations that rely on just the message.
        super(GrafanaException, self).__init__(message)


class GrafanaTimeoutError(GrafanaException):
    """
    A timeout occurred
    """


class GrafanaServerError(GrafanaException):
    """
    5xx
    """

    pass


class GrafanaClientError(GrafanaException):
    """
    Invalid input (4xx errors)
    """

    pass


class GrafanaBadInputError(GrafanaClientError):
    """
    400
    """

    def __init__(self, response):
        super(GrafanaBadInputError, self).__init__(400, response, f"Bad Input: `{response}`")


class GrafanaUnauthorizedError(GrafanaClientError):
    """
    401
    """

    def __init__(self, response):
        super(GrafanaUnauthorizedError, self).__init__(401, response, "Unauthorized")


class TokenAuth(niquests.auth.AuthBase):
    def __init__(self, token):
        self.token = token

    def __call__(self, request):
        request.headers.update({"Authorization": f"Bearer {self.token}"})
        return request


class HeaderAuth(niquests.auth.AuthBase):
    def __init__(self, name, value):
        self.name = name
        self.value = value

    def __call__(self, request):
        request.headers.update({self.name: self.value})
        return request


class GrafanaClient:
    def __init__(
        self,
        auth,
        host="localhost",
        port=None,
        url_path_prefix="",
        protocol="http",
        verify=True,
        timeout=DEFAULT_TIMEOUT,
        user_agent: str = None,
        organization_id: int = None,
        session_pool_size=DEFAULT_SESSION_POOL_SIZE,
    ):
        self.auth = auth
        self.verify = verify
        self.timeout = timeout
        self.url_host = host
        self.url_port = port
        self.url_path_prefix = url_path_prefix
        self.url_protocol = protocol
        self.session_pool_size = session_pool_size

        def construct_api_url():
            params = {
                "protocol": self.url_protocol,
                "host": self.url_host,
                "url_path_prefix": self.url_path_prefix,
            }

            if self.url_port is None:
                url_pattern = "{protocol}://{host}/{url_path_prefix}api"
            else:
                params["port"] = self.url_port
                url_pattern = "{protocol}://{host}:{port}/{url_path_prefix}api"

            return url_pattern.format(**params)

        self.url = construct_api_url()

        from grafana_client import __appname__, __version__

        self.user_agent = user_agent or f"{__appname__}/{__version__}"

        self.s = niquests.Session(pool_maxsize=session_pool_size)
        self.s.headers["User-Agent"] = self.user_agent

        self.organization_id = organization_id
        if self.organization_id:
            # orgId is defined in the openapi3 spec as an int64, but headers need to be a str
            self.s.headers["X-Grafana-Org-Id"] = str(self.organization_id)

        if self.auth is not None:
            if isinstance(self.auth, niquests.auth.AuthBase):
                pass
            elif isinstance(self.auth, tuple):
                self.auth = niquests.auth.HTTPBasicAuth(*self.auth)
            else:
                self.auth = TokenAuth(self.auth)

    def _make_url(self, url):
        return f"{self.url}{url}"

    @staticmethod
    def _ensure_valid_json_arg(json):
        if json is not None and not isinstance(json, (dict, list)):
            raise TypeError(
                f"JSON request payload has invalid shape. "
                f"Accepted are dictionaries and lists. "
                f"The type is: {type(json)}"
            )

    @staticmethod
    def _extract_from_response(r, accept_empty_json):
        if r.status_code >= 400:
            try:
                response = r.json()
            except ValueError:
                response = r.text
            message = response["message"] if isinstance(response, dict) and "message" in response else r.text

            if 500 <= r.status_code < 600:
                raise GrafanaServerError(
                    r.status_code,
                    response,
                    f"Server Error {r.status_code}: {message}",
                )
            elif r.status_code == 400:
                raise GrafanaBadInputError(response)
            elif r.status_code == 401:
                raise GrafanaUnauthorizedError(response)
            elif 400 <= r.status_code < 500:
                raise GrafanaClientError(
                    r.status_code,
                    response,
                    f"Client Error {r.status_code}: {message}",
                )

        # `204 No Content` responses have an empty response body,
        # so it doesn't decode well from JSON.
        if r.status_code == 204:
            return None

        # The "Tempo" data source responds with text/plain.
        content_type = r.headers.get("Content-Type", "")
        if content_type.startswith("text/"):
            return r.text
        try:
            return r.json()
        except JSONDecodeError:
            if accept_empty_json and r.text == "":
                return ""
            else:
                raise

    def __getattr__(self, item):
        def __request_runner(url, json=None, data=None, params=None, headers=None, accept_empty_json=False):
            __url = self._make_url(url)
            # Sanity checks.
            self._ensure_valid_json_arg(json)

            try:
                r = self.s.request(
                    item.lower(),
                    __url,
                    json=json,
                    data=data,
                    params=params,
                    headers=headers,
                    auth=self.auth,
                    verify=self.verify,
                    timeout=self.timeout,
                )
            except Timeout as e:
                raise GrafanaTimeoutError(0, None, str(e)) from e
            except HTTPError as e:
                # Make sure to not leak any exception types of the requests implementation.
                raise GrafanaException(0, None, str(e)) from e

            return self._extract_from_response(r, accept_empty_json)

        return __request_runner


class AsyncGrafanaClient(GrafanaClient):
    def __init__(
        self,
        auth,
        host="localhost",
        port=None,
        url_path_prefix="",
        protocol="http",
        verify=True,
        timeout=DEFAULT_TIMEOUT,
        user_agent: str = None,
        organization_id: int = None,
        session_pool_size=DEFAULT_SESSION_POOL_SIZE,
    ):
        super().__init__(
            auth,
            host=host,
            port=port,
            url_path_prefix=url_path_prefix,
            protocol=protocol,
            verify=verify,
            timeout=timeout,
            user_agent=user_agent,
            organization_id=organization_id,
            session_pool_size=session_pool_size,
        )
        self.s = niquests.AsyncSession(pool_maxsize=session_pool_size)
        self.s.headers.setdefault("Connection", "keep-alive")

    def __getattr__(self, item):
        async def __request_runner(url, json=None, data=None, params=None, headers=None, accept_empty_json=False):
            __url = self._make_url(url)
            # Sanity checks.
            self._ensure_valid_json_arg(json)

            try:
                r = await self.s.request(
                    item.lower(),
                    __url,
                    json=json,
                    data=data,
                    params=params,
                    headers=headers,
                    auth=self.auth,
                    verify=self.verify,
                    timeout=self.timeout,
                )
            except Timeout as e:
                raise GrafanaTimeoutError(0, None, str(e)) from e
            except HTTPError as e:
                raise GrafanaException(0, None, str(e)) from e

            return self._extract_from_response(r, accept_empty_json)

        return __request_runner



from .base import Base


class Folder(Base):
    def __init__(self, client):
        super(Folder, self).__init__(client)
        self.client = client

    def get_all_folders(self, parent_uid=None):
        """

        :return:
        """
        path = "/folders"
        params = {}
        if parent_uid:
            params["parentUid"] = parent_uid
        return self.client.GET(path, params=params)

    def get_folder(self, uid):
        """

        :param uid:
        :return:
        """
        path = "/folders/%s" % uid
        return self.client.GET(path)

    def create_folder(self, title, uid=None, parent_uid=None):
        """

        :param title:
        :param uid:
        :param parent_uid:
        :return:
        """
        json_data = dict(title=title)
        if uid is not None:
            json_data["uid"] = uid
        if parent_uid is not None:
            json_data["parentUid"] = parent_uid
        return self.client.POST("/folders", json=json_data)

    def move_folder(self, uid, parent_uid):
        """
        Move a folder beneath another parent folder.

        This is relevant only if nested folders are enabled.

        :param uid:
        :param parent_uid:
        :return:
        """
        path = "/folders/%s/move" % uid
        return self.client.POST(path, json={"parentUid": parent_uid})

    def update_folder(self, uid, title=None, version=None, overwrite=False, new_uid=None):
        """

        :param uid:
        :param title:
        :param version:
        :param overwrite:
        :param new_uid:
        :return:
        """
        body = {}
        if new_uid:
            body["uid"] = new_uid
        if title:
            body["title"] = title
        if version:
            body["version"] = version
        if overwrite:
            body["overwrite"] = True

        path = "/folders/%s" % uid
        return self.client.PUT(path, json=body)

    def delete_folder(self, uid):
        """

        :param uid:
        :return:
        """
        path = "/folders/%s" % uid
        return self.client.DELETE(path)

    def get_folder_by_id(self, folder_id):
        """

        :param folder_id:
        :return:
        """
        path = "/folders/id/%s" % folder_id
        return self.client.GET(path)

    def get_folder_permissions(self, uid):
        """

        :return:
        """
        path = "/folders/%s/permissions" % uid
        return self.client.GET(path)

    def update_folder_permissions(self, uid, items):
        """

        :param uid:
        :param items:
        :return:
        """
        update_folder_permissions_path = "/folders/%s/permissions" % uid
        return self.client.POST(update_folder_permissions_path, json=items)

    def update_folder_permissions_for_user(self, uid, user_id, items):
        """

        :param uid:
        :param user_id:
        :param items:
            {"permission": "View"} or {"permission": "Edit"} or {"permission": ""}
        :return:
        """

        update_folder_permissions_path_for_user = "/access-control/folders/%s/users/%s" % (uid, user_id)
        return self.client.POST(update_folder_permissions_path_for_user, json=items)



RBAC_ROLES_ALL = [
    {
        "version": 5,
        "uid": "vi9mlLjGz",
        "name": "fixed:datasources.permissions:writer",
        "description": "Create, read or delete data source permissions.",
        "global": True,
        "updated": "2021-05-13T22:41:49+02:00",
        "created": "2021-05-13T16:24:26+02:00",
    }
]

PERMISSION_RBAC_DATASOURCE = [
    {
        "id": 1,
        "roleName": "fixed:datasources:reader",
        "isManaged": False,
        "isInherited": False,
        "isServiceAccount": False,
        "userId": 1,
        "userLogin": "admin_user",
        "userAvatarUrl": "/avatar/admin_user",
        "actions": [
            "datasources:read",
            "datasources:query",
            "datasources:read",
            "datasources:query",
            "datasources:write",
            "datasources:delete",
        ],
        "permission": "Edit",
    },
    {
        "id": 2,
        "roleName": "managed:teams:1:permissions",
        "isManaged": True,
        "isInherited": False,
        "isServiceAccount": False,
        "team": "A team",
        "teamId": 1,
        "teamAvatarUrl": "/avatar/523d70c8551046f441727d690431858c",
        "actions": ["datasources:read", "datasources:query"],
        "permission": "Query",
    },
    {
        "id": 3,
        "roleName": "basic:admin",
        "isManaged": False,
        "isInherited": False,
        "isServiceAccount": False,
        "builtInRole": "Admin",
        "actions": ["datasources:query", "datasources:read", "datasources:write", "datasources:delete"],
        "permission": "Edit",
    },
]



import unittest

import requests
import requests_mock

from grafana_client import GrafanaApi
from grafana_client.client import GrafanaBadInputError, GrafanaClientError


class LibraryElementTestCase(unittest.TestCase):
    ConnectedPanelJSON: dict = {
        "id": 2,
        "kind": 1,
        "meta": {"connectedDashboards": 2, "folderName": "Pulsar", "folderUid": "d6818acd-f7b1-433e-a679-7f206a7ce37a"},
        "model": {
            "datasource": {"type": "prometheus", "uid": "d2caab40-4055-4236-a9b3-67ae334e096c"},
            "gridPos": {"h": 8, "w": 12, "x": 0, "y": 1},
            "id": 1,
            "libraryPanel": {"name": "CPU Seconds", "uid": "cec85d6f-834b-427e-8993-562d34fff5c4"},
            "targets": [
                {
                    "datasource": {"type": "prometheus", "uid": "d2caab40-4055-4236-a9b3-67ae334e096c"},
                    "disableTextWrap": False,
                    "editorMode": "builder",
                    "expr": 'rate(process_cpu_seconds_total{job=~"$Job", instance=~"$Instance"}[$__rate_interval])',
                    "fullMetaSearch": False,
                    "includeNullMetadata": True,
                    "instant": False,
                    "legendFormat": "__auto",
                    "range": True,
                    "refId": "A",
                    "useBackend": False,
                }
            ],
            "title": "CPU Seconds",
            "type": "timeseries",
        },
        "name": "CPU Seconds",
        "orgId": 1,
        "type": "timeseries",
        "uid": "cec85d6f-834b-427e-8993-562d34fff5c4",
        "version": 1,
    }
    ConnectedPanelUID: str = "cec85d6f-834b-427e-8993-562d34fff5c4"
    ConnectedPanelName: str = "CPU Seconds"
    ConnectedPanelConnectionsJSON: dict = {
        "result": [
            {
                "id": 2,
                "kind": 1,
                "elementId": 2,
                "connectionId": 101,
                "connectionUid": "de3791ac-6079-4c18-bde0-cb390c079722",
                "created": "2024-02-06T12:19:14-06:00",
                "createdBy": {"avatarUrl": "/avatar/46d229b033af06a191ff2267bca9ae56", "id": 1, "name": "admin"},
            },
            {
                "id": 5,
                "kind": 1,
                "elementId": 2,
                "connectionId": 102,
                "connectionUid": "a45fbfd0-b211-45fc-96ae-a56886075948",
                "created": "2024-02-06T13:21:12-06:00",
                "createdBy": {"avatarUrl": "/avatar/46d229b033af06a191ff2267bca9ae56", "id": 1, "name": "admin"},
            },
        ]
    }
    ConnectedPanelConnectionUIDs: list = (
        "de3791ac-6079-4c18-bde0-cb390c079722",
        "a45fbfd0-b211-45fc-96ae-a56886075948",
    )
    UnconnectedPanelJSON: dict = {
        "id": 3,
        "kind": 1,
        "meta": {
            "folderName": "Pulsar",
            "folderUid": "d6818acd-f7b1-433e-a679-7f206a7ce37a",
        },
        "model": {
            "datasource": {"type": "prometheus", "uid": "fb5e0357-258c-4831-b447-565be35828b5"},
            "gridPos": {"h": 8, "w": 12, "x": 12, "y": 0},
            "id": 3,
            "libraryPanel": {"name": "Heap Memory", "uid": "b07d36c0-b5f6-4228-b0c0-d3c21e16a5f6"},
            "targets": [
                {
                    "datasource": {"type": "prometheus", "uid": "fb5e0357-258c-4831-b447-565be35828b5"},
                    "disableTextWrap": False,
                    "editorMode": "builder",
                    "expr": 'jvm_memory_bytes_committed{job=~"$Job", instance=~"$Instance", area="heap"}',
                    "fullMetaSearch": False,
                    "includeNullMetadata": True,
                    "instant": False,
                    "legendFormat": "Committed",
                    "range": True,
                    "refId": "A",
                    "useBackend": False,
                },
                {
                    "datasource": {"type": "prometheus", "uid": "fb5e0357-258c-4831-b447-565be35828b5"},
                    "disableTextWrap": False,
                    "editorMode": "builder",
                    "expr": 'jvm_memory_bytes_used{job=~"$Job", instance=~"$Instance", area="heap"}',
                    "fullMetaSearch": False,
                    "hide": False,
                    "includeNullMetadata": True,
                    "instant": False,
                    "legendFormat": "Used",
                    "range": True,
                    "refId": "B",
                    "useBackend": False,
                },
                {
                    "datasource": {"type": "prometheus", "uid": "fb5e0357-258c-4831-b447-565be35828b5"},
                    "disableTextWrap": False,
                    "editorMode": "builder",
                    "expr": 'jvm_memory_bytes_max{job=~"$Job", instance=~"$Instance", area="heap"}',
                    "fullMetaSearch": False,
                    "hide": False,
                    "includeNullMetadata": True,
                    "instant": False,
                    "legendFormat": "Max",
                    "range": True,
                    "refId": "C",
                    "useBackend": False,
                },
            ],
            "title": "Heap Memory",
            "type": "timeseries",
        },
        "name": "Heap Memory",
        "orgId": 1,
        "type": "timeseries",
        "uid": "b07d36c0-b5f6-4228-b0c0-d3c21e16a5f6",
        "version": 1,
    }
    UnconnectedPanelUID: str = "b07d36c0-b5f6-4228-b0c0-d3c21e16a5f6"
    UnconnectedPanelName: str = "Heap Memory"
    UnconnectedPanelConnectionsJSON: dict = {"result": []}
    MissingPanelUID: str = "missing-panel"
    MissingPanelName: str = "Unknown name"

    CreatePanelModelJSON: dict = {
        "datasource": {"type": "prometheus", "uid": "d2caab40-4055-4236-a9b3-67ae334e096c"},
        "description": "",
        "fieldConfig": {
            "defaults": {
                "color": {"mode": "palette-classic"},
                "custom": {
                    "axisBorderShow": False,
                    "axisCenteredZero": False,
                    "axisColorMode": "text",
                    "axisLabel": "",
                    "axisPlacement": "auto",
                    "barAlignment": 0,
                    "drawStyle": "line",
                    "fillOpacity": 0,
                    "gradientMode": "none",
                    "hideFrom": {"legend": False, "tooltip": False, "viz": False},
                    "insertNulls": False,
                    "lineInterpolation": "linear",
                    "lineWidth": 1,
                    "pointSize": 5,
                    "scaleDistribution": {"type": "linear"},
                    "showPoints": "auto",
                    "spanNulls": False,
                    "stacking": {"group": "A", "mode": "none"},
                    "thresholdsStyle": {"mode": "off"},
                },
                "mappings": [],
                "thresholds": {
                    "mode": "absolute",
                    "steps": [{"color": "green", "value": None}, {"color": "red", "value": 80}],
                },
                "unit": "s",
            },
            "overrides": [],
        },
        "gridPos": {"h": 8, "w": 12, "x": 0, "y": 1},
        "libraryPanel": {"name": "CPU Seconds", "uid": "cec85d6f-834b-427e-8993-562d34fff5c4"},
        "options": {
            "legend": {"calcs": [], "displayMode": "list", "placement": "bottom", "showLegend": False},
            "tooltip": {"mode": "multi", "sort": "none"},
        },
        "targets": [
            {
                "datasource": {"type": "prometheus", "uid": "d2caab40-4055-4236-a9b3-67ae334e096c"},
                "disableTextWrap": False,
                "editorMode": "builder",
                "expr": 'rate(process_cpu_seconds_total{job=~"$Job", instance=~"$Instance"}[$__rate_interval])',
                "fullMetaSearch": False,
                "includeNullMetadata": True,
                "instant": False,
                "legendFormat": "__auto",
                "range": True,
                "refId": "A",
                "useBackend": False,
            }
        ],
        "title": "CPU Seconds",
        "transparent": True,
        "type": "timeseries",
    }
    CreatePanelUID: str = "cec85d6f-834b-427e-8993-562d34fff5c4"
    CreatePanelName: str = "CPU Seconds"
    CreatePanelFolderUID: str = "d6818acd-f7b1-433e-a679-7f206a7ce37a"
    CreatePanelResponseJSON: dict = {
        "result": {
            "id": 4,
            "orgId": 1,
            "folderId": 100,
            "folderUid": "d6818acd-f7b1-433e-a679-7f206a7ce37a",
            "uid": "cec85d6f-834b-427e-8993-562d34fff5c4",
            "name": "CPU Seconds",
            "kind": 1,
            "type": "timeseries",
            "description": "",
            "model": {
                "datasource": {"type": "prometheus", "uid": "d2caab40-4055-4236-a9b3-67ae334e096c"},
                "description": "",
                "fieldConfig": {
                    "defaults": {
                        "color": {"mode": "palette-classic"},
                        "custom": {
                            "axisBorderShow": False,
                            "axisCenteredZero": False,
                            "axisColorMode": "text",
                            "axisLabel": "",
                            "axisPlacement": "auto",
                            "barAlignment": 0,
                            "drawStyle": "line",
                            "fillOpacity": 0,
                            "gradientMode": "none",
                            "hideFrom": {"legend": False, "tooltip": False, "viz": False},
                            "insertNulls": False,
                            "lineInterpolation": "linear",
                            "lineWidth": 1,
                            "pointSize": 5,
                            "scaleDistribution": {"type": "linear"},
                            "showPoints": "auto",
                            "spanNulls": False,
                            "stacking": {"group": "A", "mode": "none"},
                            "thresholdsStyle": {"mode": "off"},
                        },
                        "mappings": [],
                        "thresholds": {
                            "mode": "absolute",
                            "steps": [{"color": "green", "value": None}, {"color": "red", "value": 80}],
                        },
                        "unit": "s",
                    },
                    "overrides": [],
                },
                "gridPos": {"h": 8, "w": 12, "x": 0, "y": 1},
                "libraryPanel": {"name": "CPU Seconds", "uid": "cec85d6f-834b-427e-8993-562d34fff5c4"},
                "options": {
                    "legend": {"calcs": [], "displayMode": "list", "placement": "bottom", "showLegend": False},
                    "tooltip": {"mode": "multi", "sort": "none"},
                },
                "targets": [
                    {
                        "datasource": {"type": "prometheus", "uid": "d2caab40-4055-4236-a9b3-67ae334e096c"},
                        "disableTextWrap": False,
                        "editorMode": "builder",
                        "expr": 'rate(process_cpu_seconds_total{job=~"$Job", instance=~"$Instance"}[$__rate_interval])',
                        "fullMetaSearch": False,
                        "includeNullMetadata": True,
                        "instant": False,
                        "legendFormat": "__auto",
                        "range": True,
                        "refId": "A",
                        "useBackend": False,
                    }
                ],
                "title": "CPU Seconds",
                "transparent": True,
                "type": "timeseries",
            },
            "version": 1,
            "meta": {
                "folderName": "Pulsar",
                "folderUid": "d6818acd-f7b1-433e-a679-7f206a7ce37a",
                "connectedDashboards": 0,
                "created": "2024-02-06T14:47:17.635073-06:00",
                "updated": "2024-02-06T14:47:17.635073-06:00",
                "createdBy": {"avatarUrl": "/avatar/46d229b033af06a191ff2267bca9ae56", "id": 1, "name": "admin"},
                "updatedBy": {"avatarUrl": "/avatar/46d229b033af06a191ff2267bca9ae56", "id": 1, "name": "admin"},
            },
        }
    }
    UpdatePanelModelJSON: dict = {
        "datasource": {"type": "prometheus", "uid": "d2caab40-4055-4236-a9b3-67ae334e096c"},
        "gridPos": {"h": 8, "w": 12, "x": 10, "y": 15},
        "id": 1,
        "libraryPanel": {"name": "CPU Seconds", "uid": "cec85d6f-834b-427e-8993-562d34fff5c4"},
        "targets": [
            {
                "datasource": {"type": "prometheus", "uid": "d2caab40-4055-4236-a9b3-67ae334e096c"},
                "disableTextWrap": False,
                "editorMode": "builder",
                "expr": 'rate(process_cpu_seconds_total{job=~"$Job", instance=~"$Instance"}[$__rate_interval])',
                "fullMetaSearch": False,
                "includeNullMetadata": True,
                "instant": False,
                "legendFormat": "__auto",
                "range": True,
                "refId": "A",
                "useBackend": False,
            }
        ],
        "title": "CPU Seconds",
        "type": "timeseries",
    }
    UpdatePanelResponseJSON: dict = {
        "result": {
            "id": 4,
            "orgId": 1,
            "folderId": 100,
            "folderUid": "d6818acd-f7b1-433e-a679-7f206a7ce37a",
            "uid": "cec85d6f-834b-427e-8993-562d34fff5c4",
            "name": "CPU Seconds",
            "kind": 1,
            "type": "timeseries",
            "description": "",
            "model": {
                "datasource": {"type": "prometheus", "uid": "d2caab40-4055-4236-a9b3-67ae334e096c"},
                "description": "",
                "gridPos": {"h": 8, "w": 12, "x": 10, "y": 15},
                "id": 2,
                "libraryPanel": {"name": "CPU Seconds", "uid": "cec85d6f-834b-427e-8993-562d34fff5c4"},
                "targets": [
                    {
                        "datasource": {"type": "prometheus", "uid": "d2caab40-4055-4236-a9b3-67ae334e096c"},
                        "disableTextWrap": False,
                        "editorMode": "builder",
                        "expr": 'rate(process_cpu_seconds_total{job=~"$Job", instance=~"$Instance"}[$__rate_interval])',
                        "fullMetaSearch": False,
                        "includeNullMetadata": True,
                        "instant": False,
                        "legendFormat": "__auto",
                        "range": True,
                        "refId": "A",
                        "useBackend": False,
                    }
                ],
                "title": "CPU Seconds",
                "type": "timeseries",
            },
            "version": 2,
            "meta": {
                "folderName": "Pulsar",
                "folderUid": "d6818acd-f7b1-433e-a679-7f206a7ce37a",
                "connectedDashboards": 0,
                "created": "2024-02-06T14:47:17-06:00",
                "updated": "2024-02-06T16:45:47.503116-06:00",
                "createdBy": {"avatarUrl": "/avatar/46d229b033af06a191ff2267bca9ae56", "id": 1, "name": "admin"},
                "updatedBy": {"avatarUrl": "/avatar/46d229b033af06a191ff2267bca9ae56", "id": 1, "name": "admin"},
            },
        }
    }
    UpdatePanelUID: str = "cec85d6f-834b-427e-8993-562d34fff5c4"
    UpdatePanelName: str = "CPU Seconds"
    UpdatePanelFolderUID: str = "d6818acd-f7b1-433e-a679-7f206a7ce37a"
    ListLibraryElementsResponseJSON: dict = {
        "result": {
            "totalCount": 2,
            "elements": [
                {
                    "id": 5,
                    "orgId": 1,
                    "folderId": 100,
                    "folderUid": "d6818acd-f7b1-433e-a679-7f206a7ce37a",
                    "uid": "cec85d6f-834b-427e-8993-562d34fff5c4",
                    "name": "CPU Seconds",
                    "kind": 1,
                    "type": "timeseries",
                    "description": "",
                    "model": {
                        "datasource": {"type": "prometheus", "uid": "d2caab40-4055-4236-a9b3-67ae334e096c"},
                        "description": "",
                        "fieldConfig": {
                            "defaults": {
                                "color": {"mode": "palette-classic"},
                                "custom": {
                                    "axisBorderShow": False,
                                    "axisCenteredZero": False,
                                    "axisColorMode": "text",
                                    "axisLabel": "",
                                    "axisPlacement": "auto",
                                    "barAlignment": 0,
                                    "drawStyle": "line",
                                    "fillOpacity": 0,
                                    "gradientMode": "none",
                                    "hideFrom": {"legend": False, "tooltip": False, "viz": False},
                                    "insertNulls": False,
                                    "lineInterpolation": "linear",
                                    "lineWidth": 1,
                                    "pointSize": 5,
                                    "scaleDistribution": {"type": "linear"},
                                    "showPoints": "auto",
                                    "spanNulls": False,
                                    "stacking": {"group": "A", "mode": "none"},
                                    "thresholdsStyle": {"mode": "off"},
                                },
                                "mappings": [],
                                "thresholds": {
                                    "mode": "absolute",
                                    "steps": [{"color": "green", "value": None}, {"color": "red", "value": 80}],
                                },
                                "unit": "s",
                            },
                            "overrides": [],
                        },
                        "gridPos": {"h": 8, "w": 12, "x": 0, "y": 1},
                        "id": 1,
                        "libraryPanel": {"name": "CPU Seconds", "uid": "cec85d6f-834b-427e-8993-562d34fff5c4"},
                        "options": {
                            "legend": {"calcs": [], "displayMode": "list", "placement": "bottom", "showLegend": False},
                            "tooltip": {"mode": "multi", "sort": "none"},
                        },
                        "targets": [
                            {
                                "datasource": {"type": "prometheus", "uid": "d2caab40-4055-4236-a9b3-67ae334e096c"},
                                "disableTextWrap": False,
                                "editorMode": "builder",
                                "expr": 'rate(process_cpu_seconds_total{job=~"$Job", instance=~"$Instance"}[$__rate_interval])',  # noqa: E501
                                "fullMetaSearch": False,
                                "includeNullMetadata": True,
                                "instant": False,
                                "legendFormat": "__auto",
                                "range": True,
                                "refId": "A",
                                "useBackend": False,
                            }
                        ],
                        "title": "CPU Seconds",
                        "transparent": True,
                        "type": "timeseries",
                    },
                    "version": 1,
                    "meta": {
                        "folderName": "Pulsar",
                        "folderUid": "d6818acd-f7b1-433e-a679-7f206a7ce37a",
                        "connectedDashboards": 0,
                        "created": "2024-02-07T10:25:19-06:00",
                        "updated": "2024-02-07T10:25:19-06:00",
                        "createdBy": {
                            "avatarUrl": "/avatar/3dbccb5e89c37491dae50c6a4be200c6",
                            "id": 2,
                            "name": "sa-tibco-messaging-monitoring",
                        },
                        "updatedBy": {
                            "avatarUrl": "/avatar/3dbccb5e89c37491dae50c6a4be200c6",
                            "id": 2,
                            "name": "sa-tibco-messaging-monitoring",
                        },
                    },
                },
                {
                    "id": 6,
                    "orgId": 1,
                    "folderId": 100,
                    "folderUid": "d6818acd-f7b1-433e-a679-7f206a7ce37a",
                    "uid": "b07d36c0-b5f6-4228-b0c0-d3c21e16a5f6",
                    "name": "Heap Memory",
                    "kind": 1,
                    "type": "timeseries",
                    "description": "",
                    "model": {
                        "datasource": {"type": "prometheus", "uid": "fb5e0357-258c-4831-b447-565be35828b5"},
                        "description": "",
                        "fieldConfig": {
                            "defaults": {
                                "color": {"mode": "palette-classic"},
                                "custom": {
                                    "axisBorderShow": False,
                                    "axisCenteredZero": False,
                                    "axisColorMode": "text",
                                    "axisLabel": "",
                                    "axisPlacement": "auto",
                                    "barAlignment": 0,
                                    "drawStyle": "line",
                                    "fillOpacity": 10,
                                    "gradientMode": "none",
                                    "hideFrom": {"legend": False, "tooltip": False, "viz": False},
                                    "insertNulls": False,
                                    "lineInterpolation": "linear",
                                    "lineWidth": 1,
                                    "pointSize": 5,
                                    "scaleDistribution": {"type": "linear"},
                                    "showPoints": "never",
                                    "spanNulls": False,
                                    "stacking": {"group": "A", "mode": "none"},
                                    "thresholdsStyle": {"mode": "off"},
                                },
                                "mappings": [],
                                "min": 0,
                                "thresholds": {
                                    "mode": "absolute",
                                    "steps": [{"color": "green", "value": None}, {"color": "red", "value": 80}],
                                },
                                "unit": "decbytes",
                            },
                            "overrides": [],
                        },
                        "gridPos": {"h": 8, "w": 12, "x": 12, "y": 0},
                        "id": 3,
                        "interval": "1m",
                        "libraryPanel": {"name": "Heap Memory", "uid": "b07d36c0-b5f6-4228-b0c0-d3c21e16a5f6"},
                        "links": [],
                        "options": {
                            "legend": {
                                "calcs": ["lastNotNull"],
                                "displayMode": "list",
                                "placement": "bottom",
                                "showLegend": False,
                            },
                            "tooltip": {"mode": "multi", "sort": "none"},
                        },
                        "pluginVersion": "10.2.2",
                        "targets": [
                            {
                                "datasource": {"type": "prometheus", "uid": "fb5e0357-258c-4831-b447-565be35828b5"},
                                "disableTextWrap": False,
                                "editorMode": "builder",
                                "expr": 'jvm_memory_bytes_committed{job=~"$Job", instance=~"$Instance", area="heap"}',
                                "fullMetaSearch": False,
                                "includeNullMetadata": True,
                                "instant": False,
                                "legendFormat": "Committed",
                                "range": True,
                                "refId": "A",
                                "useBackend": False,
                            },
                            {
                                "datasource": {"type": "prometheus", "uid": "fb5e0357-258c-4831-b447-565be35828b5"},
                                "disableTextWrap": False,
                                "editorMode": "builder",
                                "expr": 'jvm_memory_bytes_used{job=~"$Job", instance=~"$Instance", area="heap"}',
                                "fullMetaSearch": False,
                                "hide": False,
                                "includeNullMetadata": True,
                                "instant": False,
                                "legendFormat": "Used",
                                "range": True,
                                "refId": "B",
                                "useBackend": False,
                            },
                            {
                                "datasource": {"type": "prometheus", "uid": "fb5e0357-258c-4831-b447-565be35828b5"},
                                "disableTextWrap": False,
                                "editorMode": "builder",
                                "expr": 'jvm_memory_bytes_max{job=~"$Job", instance=~"$Instance", area="heap"}',
                                "fullMetaSearch": False,
                                "hide": False,
                                "includeNullMetadata": True,
                                "instant": False,
                                "legendFormat": "Max",
                                "range": True,
                                "refId": "C",
                                "useBackend": False,
                            },
                        ],
                        "title": "Heap Memory",
                        "transparent": True,
                        "type": "timeseries",
                    },
                    "version": 1,
                    "meta": {
                        "folderName": "Pulsar",
                        "folderUid": "d6818acd-f7b1-433e-a679-7f206a7ce37a",
                        "connectedDashboards": 0,
                        "created": "2024-02-07T10:25:37-06:00",
                        "updated": "2024-02-07T10:25:37-06:00",
                        "createdBy": {
                            "avatarUrl": "/avatar/3dbccb5e89c37491dae50c6a4be200c6",
                            "id": 2,
                            "name": "sa-tibco-messaging-monitoring",
                        },
                        "updatedBy": {
                            "avatarUrl": "/avatar/3dbccb5e89c37491dae50c6a4be200c6",
                            "id": 2,
                            "name": "sa-tibco-messaging-monitoring",
                        },
                    },
                },
            ],
            "page": 1,
            "perPage": 100,
        }
    }

    HealthResponsePre8_2: dict = {"commit": "unknown", "database": "ok", "version": "8.1.8"}
    HealthResponsePost8_2: dict = {"commit": "unknown-dev", "database": "ok", "version": "10.2.2"}

    def setUp(self):
        self.grafana = GrafanaApi(("admin", "admin"), host="localhost", url_path_prefix="", protocol="http")

    @requests_mock.Mocker()
    def test_get_library_element(self, m):
        m.get(
            "http://localhost/api/health",
            json=LibraryElementTestCase.HealthResponsePost8_2,
        )
        m.get(
            f"http://localhost/api/library-elements/{LibraryElementTestCase.ConnectedPanelUID}",
            json=LibraryElementTestCase.ConnectedPanelJSON,
        )
        element = self.grafana.libraryelement.get_library_element(LibraryElementTestCase.ConnectedPanelUID)
        self.assertEqual(element["uid"], LibraryElementTestCase.ConnectedPanelUID)

    @requests_mock.Mocker()
    def test_get_library_element_grafana_pre_8_2(self, m):
        m.get(
            "http://localhost/api/health",
            json=LibraryElementTestCase.HealthResponsePre8_2,
        )
        m.get(
            f"http://localhost/api/library-elements/{LibraryElementTestCase.ConnectedPanelUID}",
            json=LibraryElementTestCase.ConnectedPanelJSON,
        )
        with self.assertRaises(DeprecationWarning) as ex:
            self.grafana.libraryelement.get_library_element(LibraryElementTestCase.ConnectedPanelUID)
        self.assertEqual(
            "Grafana versions earlier than 8.2 do not support library elements",
            str(ex.exception),
        )

    @requests_mock.Mocker()
    def test_get_library_element_notfound(self, m):
        m.get(
            "http://localhost/api/health",
            json=LibraryElementTestCase.HealthResponsePost8_2,
        )
        m.get(
            f"http://localhost/api/library-elements/{LibraryElementTestCase.MissingPanelUID}",
            json={"message": "library element could not be found"},
            status_code=404,
        )
        with self.assertRaises(GrafanaClientError) as ex:
            self.grafana.libraryelement.get_library_element(LibraryElementTestCase.MissingPanelUID)
        self.assertEqual(404, ex.exception.status_code)
        self.assertEqual("Client Error 404: library element could not be found", ex.exception.message)

    @requests_mock.Mocker()
    def test_get_library_element_notfound_grafana_pre_8_2(self, m):
        def custom_matcher(request):
            if request.path_url == "/api/health":
                return None
            resp = requests.Response()
            resp.status_code = 404
            resp._content = b'{"message": "library element could not be found"}'
            resp.headers["Content-Type"] = "application/json"
            return resp

        m.get(
            "http://localhost/api/health",
            json=LibraryElementTestCase.HealthResponsePre8_2,
        )
        m.add_matcher(custom_matcher)
        with self.assertRaises(DeprecationWarning) as ex:
            self.grafana.libraryelement.get_library_element(LibraryElementTestCase.MissingPanelUID)
        self.assertEqual(
            "Grafana versions earlier than 8.2 do not support library elements",
            str(ex.exception),
        )

    @requests_mock.Mocker()
    def test_get_library_element_by_name(self, m):
        m.get(
            "http://localhost/api/health",
            json=LibraryElementTestCase.HealthResponsePost8_2,
        )
        m.get(
            f"http://localhost/api/library-elements/name/{LibraryElementTestCase.ConnectedPanelName}",
            json=LibraryElementTestCase.ConnectedPanelJSON,
        )
        element = self.grafana.libraryelement.get_library_element_by_name(LibraryElementTestCase.ConnectedPanelName)
        self.assertEqual(element["name"], LibraryElementTestCase.ConnectedPanelName)

    @requests_mock.Mocker()
    def test_get_library_element_by_name_grafana_pre_8_2(self, m):
        m.get(
            "http://localhost/api/health",
            json=LibraryElementTestCase.HealthResponsePre8_2,
        )
        m.get(
            f"http://localhost/api/library-elements/name/{LibraryElementTestCase.ConnectedPanelName}",
            json=LibraryElementTestCase.ConnectedPanelJSON,
        )
        with self.assertRaises(DeprecationWarning) as ex:
            self.grafana.libraryelement.get_library_element_by_name(LibraryElementTestCase.ConnectedPanelName)
        self.assertEqual(
            "Grafana versions earlier than 8.2 do not support library elements",
            str(ex.exception),
        )

    @requests_mock.Mocker()
    def test_get_library_element_by_name_notfound(self, m):
        m.get(
            "http://localhost/api/health",
            json=LibraryElementTestCase.HealthResponsePost8_2,
        )
        m.get(
            f"http://localhost/api/library-elements/name/{LibraryElementTestCase.MissingPanelName}",
            json={"message": "library element could not be found"},
            status_code=404,
        )
        with self.assertRaises(GrafanaClientError) as ex:
            self.grafana.libraryelement.get_library_element_by_name(LibraryElementTestCase.MissingPanelName)
        self.assertEqual(404, ex.exception.status_code)
        self.assertEqual("Client Error 404: library element could not be found", ex.exception.message)

    @requests_mock.Mocker()
    def test_get_library_element_by_name_notfound_grafana_pre_8_2(self, m):
        def custom_matcher(request):
            if request.path_url == "/api/health":
                return None
            resp = requests.Response()
            resp.status_code = 404
            resp._content = b'{"message": "library element could not be found"}'
            resp.headers["Content-Type"] = "application/json"
            return resp

        m.get(
            "http://localhost/api/health",
            json=LibraryElementTestCase.HealthResponsePre8_2,
        )
        m.add_matcher(custom_matcher)
        with self.assertRaises(DeprecationWarning) as ex:
            self.grafana.libraryelement.get_library_element_by_name(LibraryElementTestCase.MissingPanelName)
        self.assertEqual(
            "Grafana versions earlier than 8.2 do not support library elements",
            str(ex.exception),
        )

    @requests_mock.Mocker()
    def test_get_library_element_connections(self, m):
        m.get(
            "http://localhost/api/health",
            json=LibraryElementTestCase.HealthResponsePost8_2,
        )
        m.get(
            f"http://localhost/api/library-elements/{LibraryElementTestCase.ConnectedPanelUID}/connections",
            json=LibraryElementTestCase.ConnectedPanelConnectionsJSON,
        )
        connections = self.grafana.libraryelement.get_library_element_connections(
            LibraryElementTestCase.ConnectedPanelUID
        )
        self.assertEqual(len(connections["result"]), 2)
        self.assertIn(connections["result"][0]["connectionUid"], LibraryElementTestCase.ConnectedPanelConnectionUIDs)
        self.assertIn(connections["result"][1]["connectionUid"], LibraryElementTestCase.ConnectedPanelConnectionUIDs)

    @requests_mock.Mocker()
    def test_get_library_element_connections_grafana_pre_8_2(self, m):
        m.get(
            "http://localhost/api/health",
            json=LibraryElementTestCase.HealthResponsePre8_2,
        )
        m.get(
            f"http://localhost/api/library-elements/{LibraryElementTestCase.ConnectedPanelUID}/connections",
            json=LibraryElementTestCase.ConnectedPanelConnectionsJSON,
        )
        with self.assertRaises(DeprecationWarning) as ex:
            self.grafana.libraryelement.get_library_element_connections(LibraryElementTestCase.ConnectedPanelUID)
        self.assertEqual(
            "Grafana versions earlier than 8.2 do not support library elements",
            str(ex.exception),
        )

    @requests_mock.Mocker()
    def test_get_library_element_connections_notfound(self, m):
        m.get(
            "http://localhost/api/health",
            json=LibraryElementTestCase.HealthResponsePost8_2,
        )
        m.get(
            f"http://localhost/api/library-elements/{LibraryElementTestCase.MissingPanelUID}/connections",
            json={"message": "library element could not be found"},
            status_code=404,
        )
        with self.assertRaises(GrafanaClientError) as ex:
            self.grafana.libraryelement.get_library_element_connections(LibraryElementTestCase.MissingPanelUID)
        self.assertEqual(404, ex.exception.status_code)
        self.assertEqual("Client Error 404: library element could not be found", ex.exception.message)

    @requests_mock.Mocker()
    def test_get_library_element_connections_notfound_grafana_pre_8_2(self, m):
        def custom_matcher(request):
            if request.path_url == "/api/health":
                return None
            resp = requests.Response()
            resp.status_code = 404
            resp._content = b'{"message": "library element could not be found"}'
            resp.headers["Content-Type"] = "application/json"
            return resp

        m.get(
            "http://localhost/api/health",
            json=LibraryElementTestCase.HealthResponsePre8_2,
        )
        m.add_matcher(custom_matcher)
        with self.assertRaises(DeprecationWarning) as ex:
            self.grafana.libraryelement.get_library_element(LibraryElementTestCase.MissingPanelUID)
        self.assertEqual(
            "Grafana versions earlier than 8.2 do not support library elements",
            str(ex.exception),
        )

    @requests_mock.Mocker()
    def test_get_library_element_connections_noconnections(self, m):
        m.get(
            "http://localhost/api/health",
            json=LibraryElementTestCase.HealthResponsePost8_2,
        )
        m.get(
            f"http://localhost/api/library-elements/{LibraryElementTestCase.UnconnectedPanelUID}/connections",
            json=LibraryElementTestCase.UnconnectedPanelConnectionsJSON,
        )
        connections = self.grafana.libraryelement.get_library_element_connections(
            LibraryElementTestCase.UnconnectedPanelUID
        )
        self.assertEqual(len(connections["result"]), 0)

    @requests_mock.Mocker()
    def test_get_library_element_connections_noconnections_grafana_pre_8_2(self, m):
        m.get(
            "http://localhost/api/health",
            json=LibraryElementTestCase.HealthResponsePre8_2,
        )
        m.get(
            f"http://localhost/api/library-elements/{LibraryElementTestCase.UnconnectedPanelUID}/connections",
            json=LibraryElementTestCase.UnconnectedPanelConnectionsJSON,
        )
        with self.assertRaises(DeprecationWarning) as ex:
            self.grafana.libraryelement.get_library_element_connections(LibraryElementTestCase.UnconnectedPanelUID)
        self.assertEqual(
            "Grafana versions earlier than 8.2 do not support library elements",
            str(ex.exception),
        )

    @requests_mock.Mocker()
    def test_create_library_element(self, m):
        m.get(
            "http://localhost/api/health",
            json=LibraryElementTestCase.HealthResponsePost8_2,
        )
        m.post(
            "http://localhost/api/library-elements",
            json=LibraryElementTestCase.CreatePanelResponseJSON,
        )
        resp = self.grafana.libraryelement.create_library_element(
            model=LibraryElementTestCase.CreatePanelModelJSON,
            name=LibraryElementTestCase.CreatePanelName,
            uid=LibraryElementTestCase.CreatePanelUID,
            folder_uid=LibraryElementTestCase.CreatePanelFolderUID,
        )
        self.assertIsNotNone(resp["result"])
        self.assertIsNotNone(resp["result"]["folderUid"])
        self.assertEqual(resp["result"]["folderUid"], LibraryElementTestCase.CreatePanelFolderUID)
        self.assertIsNotNone(resp["result"]["uid"])
        self.assertEqual(resp["result"]["uid"], LibraryElementTestCase.CreatePanelUID)
        self.assertIsNotNone(resp["result"]["model"])

    @requests_mock.Mocker()
    def test_create_library_element_grafana_pre_8_2(self, m):
        m.get(
            "http://localhost/api/health",
            json=LibraryElementTestCase.HealthResponsePre8_2,
        )
        m.post(
            "http://localhost/api/library-elements",
            json=LibraryElementTestCase.CreatePanelResponseJSON,
        )
        with self.assertRaises(DeprecationWarning) as ex:
            self.grafana.libraryelement.create_library_element(
                model=LibraryElementTestCase.CreatePanelModelJSON,
                name=LibraryElementTestCase.CreatePanelName,
                uid=LibraryElementTestCase.CreatePanelUID,
                folder_uid=LibraryElementTestCase.CreatePanelFolderUID,
            )
        self.assertEqual(
            "Grafana versions earlier than 8.2 do not support library elements",
            str(ex.exception),
        )

    @requests_mock.Mocker()
    def test_create_library_element_already_exists(self, m):
        m.get(
            "http://localhost/api/health",
            json=LibraryElementTestCase.HealthResponsePost8_2,
        )
        m.post(
            "http://localhost/api/library-elements",
            json={"message": "library element with that name or UID already exists"},
            status_code=400,
        )
        with self.assertRaisesRegex(
            GrafanaBadInputError, "Bad Input: .*library element with that name or UID already exists.*"
        ):
            self.grafana.libraryelement.create_library_element(
                model=LibraryElementTestCase.CreatePanelModelJSON,
                name=LibraryElementTestCase.CreatePanelName,
                uid=LibraryElementTestCase.CreatePanelUID,
                folder_uid=LibraryElementTestCase.CreatePanelFolderUID,
            )

    @requests_mock.Mocker()
    def test_update_library_element(self, m):
        m.get(
            "http://localhost/api/health",
            json=LibraryElementTestCase.HealthResponsePost8_2,
        )
        m.patch(
            f"http://localhost/api/library-elements/{LibraryElementTestCase.UpdatePanelUID}",
            json=LibraryElementTestCase.UpdatePanelResponseJSON,
        )
        resp = self.grafana.libraryelement.update_library_element(
            model=LibraryElementTestCase.UpdatePanelModelJSON,
            name=LibraryElementTestCase.UpdatePanelName,
            uid=LibraryElementTestCase.UpdatePanelUID,
            folder_uid=LibraryElementTestCase.UpdatePanelFolderUID,
            version=1,
        )
        self.assertIsNotNone(resp["result"])
        self.assertIsNotNone(resp["result"]["folderUid"])
        self.assertEqual(resp["result"]["folderUid"], LibraryElementTestCase.CreatePanelFolderUID)
        self.assertIsNotNone(resp["result"]["uid"])
        self.assertEqual(resp["result"]["uid"], LibraryElementTestCase.CreatePanelUID)
        self.assertIsNotNone(resp["result"]["model"])
        self.assertIsNotNone(resp["result"]["model"]["gridPos"])
        self.assertIsNotNone(resp["result"]["model"]["gridPos"]["x"])
        self.assertEqual(resp["result"]["model"]["gridPos"]["x"], 10)
        self.assertIsNotNone(resp["result"]["model"]["gridPos"]["y"])
        self.assertEqual(resp["result"]["model"]["gridPos"]["y"], 15)
        self.assertIsNotNone(resp["result"]["version"])
        self.assertEqual(resp["result"]["version"], 2)

    @requests_mock.Mocker()
    def test_update_library_element_grafana_pre_8_2(self, m):
        m.get(
            "http://localhost/api/health",
            json=LibraryElementTestCase.HealthResponsePre8_2,
        )
        m.patch(
            f"http://localhost/api/library-elements/{LibraryElementTestCase.UpdatePanelUID}",
            json=LibraryElementTestCase.UpdatePanelResponseJSON,
        )
        with self.assertRaises(DeprecationWarning) as ex:
            self.grafana.libraryelement.update_library_element(
                model=LibraryElementTestCase.UpdatePanelModelJSON,
                name=LibraryElementTestCase.UpdatePanelName,
                uid=LibraryElementTestCase.UpdatePanelUID,
                folder_uid=LibraryElementTestCase.UpdatePanelFolderUID,
                version=1,
            )
        self.assertEqual(
            "Grafana versions earlier than 8.2 do not support library elements",
            str(ex.exception),
        )

    @requests_mock.Mocker()
    def test_update_library_element_notfound(self, m):
        m.get(
            "http://localhost/api/health",
            json=LibraryElementTestCase.HealthResponsePost8_2,
        )
        m.patch(
            f"http://localhost/api/library-elements/{LibraryElementTestCase.MissingPanelUID}",
            json={"message": "library element could not be found"},
            status_code=404,
        )
        with self.assertRaises(GrafanaClientError) as ex:
            self.grafana.libraryelement.update_library_element(
                model=LibraryElementTestCase.UpdatePanelModelJSON,
                name=LibraryElementTestCase.UpdatePanelName,
                uid=LibraryElementTestCase.MissingPanelUID,
                folder_uid=LibraryElementTestCase.UpdatePanelFolderUID,
                version=1,
            )
        self.assertEqual(404, ex.exception.status_code)
        self.assertEqual("Client Error 404: library element could not be found", ex.exception.message)

    @requests_mock.Mocker()
    def test_delete_library_element_connections(self, m):
        m.get(
            "http://localhost/api/health",
            json=LibraryElementTestCase.HealthResponsePost8_2,
        )
        m.delete(
            f"http://localhost/api/library-elements/{LibraryElementTestCase.ConnectedPanelUID}",
            json={"message": "the library element has connections"},
            status_code=403,
        )
        with self.assertRaises(GrafanaClientError) as ex:
            self.grafana.libraryelement.delete_library_element(LibraryElementTestCase.ConnectedPanelUID)
        self.assertEqual(403, ex.exception.status_code)
        self.assertEqual("Client Error 403: the library element has connections", ex.exception.message)

    @requests_mock.Mocker()
    def test_delete_library_element_noconnections(self, m):
        m.get(
            "http://localhost/api/health",
            json=LibraryElementTestCase.HealthResponsePost8_2,
        )
        m.delete(
            f"http://localhost/api/library-elements/{LibraryElementTestCase.UnconnectedPanelUID}",
            json={"id": 4, "message": "Library element deleted"},
        )
        resp = self.grafana.libraryelement.delete_library_element(LibraryElementTestCase.UnconnectedPanelUID)
        self.assertIsNotNone(resp)
        self.assertIsNotNone(resp.get("message"))
        self.assertEqual("Library element deleted", resp.get("message"))

    @requests_mock.Mocker()
    def test_delete_library_element_notfound(self, m):
        m.get(
            "http://localhost/api/health",
            json=LibraryElementTestCase.HealthResponsePost8_2,
        )
        m.delete(
            f"http://localhost/api/library-elements/{LibraryElementTestCase.MissingPanelUID}",
            json={"message": "library element could not be found"},
            status_code=404,
        )
        with self.assertRaises(GrafanaClientError) as ex:
            self.grafana.libraryelement.delete_library_element(LibraryElementTestCase.MissingPanelUID)
        self.assertEqual(404, ex.exception.status_code)
        self.assertEqual("Client Error 404: library element could not be found", ex.exception.message)

    @requests_mock.Mocker()
    def test_delete_library_element_notfound_grafana_pre_8_2(self, m):
        def custom_matcher(request):
            if request.path_url == "/api/health":
                return None
            resp = requests.Response()
            resp.status_code = 404
            resp._content = b'{"message": "library element could not be found"}'
            resp.headers["Content-Type"] = "application/json"
            return resp

        m.get(
            "http://localhost/api/health",
            json=LibraryElementTestCase.HealthResponsePre8_2,
        )
        m.add_matcher(custom_matcher)
        with self.assertRaises(DeprecationWarning) as ex:
            self.grafana.libraryelement.delete_library_element(LibraryElementTestCase.MissingPanelUID)
        self.assertEqual(
            "Grafana versions earlier than 8.2 do not support library elements",
            str(ex.exception),
        )

    @requests_mock.Mocker()
    def test_list_library_elements(self, m):
        m.get(
            "http://localhost/api/health",
            json=LibraryElementTestCase.HealthResponsePost8_2,
        )
        m.get(
            "http://localhost/api/library-elements?",
            json=LibraryElementTestCase.ListLibraryElementsResponseJSON,
        )
        elements = self.grafana.libraryelement.list_library_elements()
        self.assertIsNotNone(elements.get("result"))
        result = elements.get("result")
        self.assertIsNotNone(result.get("totalCount"))
        self.assertEqual(result["totalCount"], 2)
        self.assertIsNotNone(result.get("page"))
        self.assertEqual(result["page"], 1)
        self.assertIsNotNone(result.get("perPage"))
        self.assertEqual(result["perPage"], 100)

    @requests_mock.Mocker()
    def test_list_library_elements_grafana_pre_8_2(self, m):
        m.get(
            "http://localhost/api/health",
            json=LibraryElementTestCase.HealthResponsePre8_2,
        )
        m.get(
            "http://localhost/api/library-elements?",
            json=LibraryElementTestCase.ListLibraryElementsResponseJSON,
        )
        with self.assertRaises(DeprecationWarning) as ex:
            self.grafana.libraryelement.list_library_elements()
        self.assertEqual(
            "Grafana versions earlier than 8.2 do not support library elements",
            str(ex.exception),
        )



"""
About
=====

Example program for interacting with the Grafana Service Account.

Usage
=====

Make sure to adjust `GrafanaApi` options at the bottom of the
file.
And replace aaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa with the uid of the folder where you want to add the service account.

Synopsis
========
::

    source .venv/bin/activate
    python examples/service-account.py

"""

import uuid

from grafana_client import GrafanaApi
from grafana_client.client import GrafanaClientError


def run_conversation(grafana: GrafanaApi):
    # print("Grafana address")
    # print(grafana.client.url)

    # print("Health check")
    # print(grafana.health.check())

    print("Create Service Account")
    sa = {"name": "Sytem Account DEMO", "role": "Viewer"}

    # Add a new service account
    try:
        response = grafana.serviceaccount.create(sa)
        print(response)
    except GrafanaClientError as ex:
        print(f"ERROR: {ex}")

    # Add a new token to service account, you need to writedown the token['key'] to use it later
    # Genera an UUID
    uuid_to_token = uuid.uuid4()
    token_name = "%s" % (uuid_to_token)

    try:
        token = grafana.serviceaccount.create_token(response["id"], {"name": token_name})
        print(token)
    except GrafanaClientError as ex:
        print(f"ERROR: {ex}")

    # Add Viewer permissions to service account created to folder with uid aaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa
    uid_folder = "aaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa"
    permissions = {"permission": "View"}
    sa_id = response["id"]
    grafana.folder.update_folder_permissions_for_user(uid_folder, sa_id, permissions)

    # Search service account by name
    search_sa = grafana.serviceaccount.search_one(service_account_name="Sytem Account DEMO")
    print(search_sa)


if __name__ == "__main__":
    # Connect to custom Grafana instance.
    grafana = GrafanaApi.from_url(url="localhost:3000", credential=("admin", "admin"))

    run_conversation(grafana)



"""
About
=====

Example program for interacting with the Grafana Teams API.
It will need `teams:{create,read}` permissions to create a team and inquire the
list of existing teams afterwards.

Usage
=====

Make sure to adjust `GrafanaApi(auth=, host=)` options at the bottom of the
file.

Synopsis
========
::

    source .venv/bin/activate
    python examples/team.py

"""

import json

from grafana_client import GrafanaApi
from grafana_client.client import GrafanaClientError


def run_conversation(grafana: GrafanaApi):
    print("Grafana address")
    print(grafana.client.url)

    print("Health check")
    print(grafana.health.check())

    print("Create team")
    team = {"name": "Testdrive Team", "email": "email@example.org"}
    try:
        response = grafana.teams.add_team(team)
        print(response)
    except GrafanaClientError as ex:
        if ex.status_code == 409:
            print(f"Team already exists: {ex.response}")
        else:
            print(f"ERROR: {ex}")

    print("Search teams")
    try:
        response = grafana.teams.search_teams()
        print(jd(response))
    except GrafanaClientError as ex:
        print(f"ERROR: {ex}")


def jd(data):
    return json.dumps(data, indent=2)


if __name__ == "__main__":
    # Connect to custom Grafana instance.
    grafana = GrafanaApi(
        auth=("admin", "admin"),
        host="localhost:3000",
    )

    # Connect to Grafana instance of Grafana Labs fame.
    # You will see permission errors on this, as you will probably have none of
    # `teams:{create,read}` on Grafana's server, let alone without authentication.
    # grafana = GrafanaApi(host="play.grafana.org", protocol="https")

    run_conversation(grafana)



from ..base import Base


class Rbac(Base):
    def __init__(self, client):
        super(Rbac, self).__init__(client)
        self.client = client

    async def get_rbac_roles_all(self):
        """
        The Rbac is only available in Grafana Enterprise.

        :return:
        """
        roles_path = "/access-control/roles"
        return await self.client.GET(roles_path)

    async def add_rbac_role_team(self, team_id, role_uid):
        """
        The Rbac is only available in Grafana Enterprise.

        :param team_id:
        :param role_uid:
        :return:
        """
        role_team_path = "/access-control/teams/%s/roles" % team_id
        return await self.client.POST(role_team_path, json={"roleUid": role_uid})

    async def add_rbac_roles_team(self, team_id, role_uids):
        """
        The Rbac is only available in Grafana Enterprise.

        :param team_id:
        :param role_uids:
        :return:
        """
        role_team_path = "/access-control/teams/%s/roles" % team_id
        return await self.client.PUT(role_team_path, json={"roleUids": role_uids})

    async def remove_rbac_role_team(self, team_id, role_uid):
        """
        The Rbac is only available in Grafana Enterprise.

        :param team_id:
        :param role_uid:
        :return:
        """
        role_team_path = "/access-control/teams/%s/roles/%s" % (team_id, role_uid)
        return await self.client.DELETE(role_team_path)

    async def get_rbac_datasources(self, datasource_uid):
        """
        Gets all existing permissions for the data source with the given uid

        The Rbac is only available in Grafana Enterprise.
        This API works only with grafana version > 10.2.3.

        :param datasource_uid:
        :return:
        """
        datasource_path = "/access-control/datasources/%s" % datasource_uid
        return await self.client.GET(datasource_path)

    async def set_rbac_datasources_teams(self, datasource_uid, team_id, permission):
        """
        Sets team permission for the data source with the given uid.

        the values allowed for the permission argument are :
        Query, Edit, or Admin.
        To remove a permission, set the permission argument to an empty string.
        The Rbac is only available in Grafana Enterprise.
        This API works only with grafana version > 10.2.3.

        :param datasource_uid:
        :param team_id:
        :param permission:
        :return:
        """
        datasource_path = "/access-control/datasources/%s/teams/%s" % (datasource_uid, team_id)
        return await self.client.POST(datasource_path, json={"permission": permission})

    async def set_rbac_datasources_builtin_roles(self, datasource_uid, builtin_role, permission):
        """
        Sets permission for the data source with the given uid to all users who
        have the specified basic role.

        the values allowed for the builtin_role argument are :
        Admin, Editor or Viewer
        the values allowed for the permission argument are :
        Query, Edit, or Admin.
        To remove a permission, set the permission argument to an empty string.
        The Rbac is only available in Grafana Enterprise.
        This API works only with grafana version > 10.2.3.

        :param datasource_uid:
        :param builtin_role:
        :param permission:
        :return:
        """
        datasource_path = "/access-control/datasources/%s/builtInRoles/%s" % (datasource_uid, builtin_role)
        return await self.client.POST(datasource_path, json={"permission": permission})



from grafana_client.util import format_param_value

from .base import Base


class Search(Base):
    def __init__(self, client):
        super(Search, self).__init__(client)
        self.client = client

    def search_dashboards(
        self,
        query=None,
        tag=None,
        type_=None,
        dashboard_ids=None,
        dashboard_uids=None,
        folder_ids=None,
        folder_uids=None,
        starred=None,
        limit=None,
        page=None,
    ):
        """

        :param query:
        :param tag:
        :param type_:
        :param dashboard_ids:
        :param dashboard_uids:
        :param folder_ids:
        :param folder_uids:
        :param starred:
        :param limit:
        :param page:
        :return:
        """
        list_dashboard_path = "/search"
        params = {}

        if query:
            params["query"] = query

        if tag:
            params["tag"] = format_param_value(tag)

        if type_:
            params["type"] = type_

        if dashboard_ids:
            params["dashboardIds"] = format_param_value(dashboard_ids)

        if dashboard_uids:
            params["dashboardUIDs"] = format_param_value(dashboard_uids)

        if folder_ids:
            params["folderIds"] = format_param_value(folder_ids)

        if folder_uids:
            params["folderUIDs"] = format_param_value(folder_uids)

        if starred:
            params["starred"] = starred

        if limit:
            params["limit"] = limit

        if page:
            params["page"] = page

        return self.client.GET(list_dashboard_path, params=params)



import asyncio
import socket

from grafana_client import AsyncGrafanaApi
from grafana_client.client import GrafanaClientError

try:
    from unittest import IsolatedAsyncioTestCase
except ImportError:
    pass
else:

    class TestAsync(IsolatedAsyncioTestCase):
        async def test_basic_async_client(self):
            try:
                sock = socket.create_connection(("play.grafana.org", 80), timeout=1)
            except (ConnectionRefusedError, socket.gaierror, TimeoutError):
                self.skipTest("Test requires a WAN access to play.grafana.org")
            else:
                sock.close()

            async def fetch_dashboard(async_client, uid):
                return await async_client.dashboard.get_dashboard(uid)

            grafana = AsyncGrafanaApi(host="play.grafana.org")

            await grafana.connect()

            version = await grafana.version

            self.assertTrue(version != "" and "." in version)

            folders = await grafana.folder.get_all_folders()

            self.assertTrue(isinstance(folders, list))
            self.assertTrue(len(folders) >= 4)

            tasks = []

            for folder in folders[:2]:
                if folder["id"] > 0:  # someone created an entry with a negative id...
                    dashboards = await grafana.search.search_dashboards(folder_uids=[folder["uid"]])
                    for dashboard in dashboards[:4]:
                        tasks.append(fetch_dashboard(grafana, dashboard["uid"]))

            results = await asyncio.gather(*tasks)

            self.assertEqual(len(results), 4)
            self.assertTrue(all(isinstance(r, dict) and list(r.keys()) for r in results))

        async def test_exception_async_client(self):
            try:
                sock = socket.create_connection(("play.grafana.org", 80), timeout=1)
            except (ConnectionRefusedError, socket.gaierror, TimeoutError):
                self.skipTest("Test requires a WAN access to play.grafana.org")
            else:
                sock.close()

            grafana = AsyncGrafanaApi(host="play.grafana.org")

            await grafana.version

            with self.assertRaises(GrafanaClientError) as exc:
                await grafana.admin.change_user_password(0, "impossible")

            self.assertEqual(exc.exception.status_code, 403)



import logging
import sys


def setup_logging(level=logging.INFO):
    log_format = "%(asctime)-15s [%(name)-35s] %(levelname)-8s: %(message)s"
    logging.basicConfig(format=log_format, stream=sys.stderr, level=level)


def as_bool(value: str) -> bool:
    """
    Given a string value that represents True or False, returns the Boolean equivalent.
    Heavily inspired from distutils strtobool.

    From `isort`: https://github.com/PyCQA/isort/blob/5.10.1/isort/settings.py#L915-L922
    """

    if value is None:
        return False

    if isinstance(value, bool):
        return value

    _STR_BOOLEAN_MAPPING = {
        "y": True,
        "yes": True,
        "t": True,
        "on": True,
        "1": True,
        "true": True,
        "n": False,
        "no": False,
        "f": False,
        "off": False,
        "0": False,
        "false": False,
    }
    try:
        return _STR_BOOLEAN_MAPPING[value.lower()]
    except KeyError:
        raise ValueError(f"invalid truth value {value}")


def format_param_value(maybe_list):
    if isinstance(maybe_list, list):
        return ",".join([str(x) for x in maybe_list])
    else:
        return maybe_list



"""
About
=====

Example program for listing folders and getting a dashboard of a remote Grafana
instance. By default, it uses `play.grafana.org`. Please adjust to your needs.


Synopsis
========
::

    source .venv/bin/activate
    python examples/folders-dashboard.py | jq
"""

import asyncio
import json
import sys
from time import perf_counter

from grafana_client import AsyncGrafanaApi


async def fetch_dashboard(grafana, uid):
    print(f"## Dashboard with UID {uid} at play.grafana.org", file=sys.stderr)
    dashboard = await grafana.dashboard.get_dashboard(uid)
    print(json.dumps(dashboard, indent=2))


async def main():
    before = perf_counter()
    # Connect to public Grafana instance of Grafana Labs fame.
    grafana = AsyncGrafanaApi(host="play.grafana.org")

    print("## All folders on play.grafana.org", file=sys.stderr)
    folders = await grafana.folder.get_all_folders()
    print(json.dumps(folders, indent=2))

    tasks = []

    for folder in folders:
        if folder["id"] > 0:
            tasks.append(fetch_dashboard(grafana, folder["uid"]))
            if len(tasks) == 4:
                break

    await asyncio.gather(*tasks)
    print(f"## Completed in {perf_counter() - before}s", file=sys.stderr)


if __name__ == "__main__":
    asyncio.run(main())



#!/usr/bin/env python
"""
Program to automatically generate asynchronous code from the
synchronous counterpart.

Synopsis:

  # Check if generated code is up-to-date.
  python script/generate_async.py check

  # Generate code.
  python script/generate_async.py format

What does this program does:
- For each module in `grafana_client/elements/*.py`, except `base` and `__init__`:
  Inject async/await keyword in all available methods / client interactions.
- Detect modules no longer in use, and remove them.
- Generate the async top level code based on `elements/__init__.py`.
"""

import os
import re
import shutil
import subprocess
import sys
from pathlib import Path
from tempfile import TemporaryDirectory

BASE_PATH = Path(".." if Path("../grafana_client").exists() else ".")

HERE = Path.cwd().absolute()
PYPROJECT_TOML = HERE / "pyproject.toml"
SOURCE = BASE_PATH / "grafana_client" / "elements"
TARGET = BASE_PATH / "grafana_client" / "elements" / "_async"


def run(action: str):
    """
    Invoke either the `check`, or the `format` subcommand.

    :check:  Create amalgamated tree in temporary directory,
             and compare with original. This is suitable for
             running sanity checks on CI.
    :format: Generate amalgamated async code from synchronous
             reference code.
    """

    run_check = action == "check" or False
    run_format = action == "format" or False

    source = SOURCE
    target = TARGET

    if run_check:
        # Create temporary formatted exemplar for probing it,
        # compare with current state, and croak on deviations.

        # Use this code for `delete=False`.
        # TemporaryDirectory._rmtree = lambda *more, **kwargs: None  # noqa: ERA001

        with TemporaryDirectory() as tmpdir:
            target = Path(tmpdir)
            process(source, target)

            command = f"diff -x __pycache__ -x .ruff_cache -u {TARGET} {target}"
            exitcode = os.system(command)

        if exitcode == 0:
            print(msg("INFO: Async code up-to-date. Excellent."))
        else:
            print(msg("ERROR: Async code not up-to-date. Please run `poe format`."))
            sys.exit(2)

    elif run_format:
        Path(target).mkdir(exist_ok=True)
        process(source, target)

    else:
        raise ValueError("Wrong or missing action, use either `check` or `format`.")


def process(source: Path, target: Path):
    """
    Process files, from input path (source) to output path (target).
    """

    module_processed = []

    print(f"Input path:  {source}")
    print(f"Output path: {target}")

    # for module_path in glob(f"{source}/*.py"):
    for module_path in Path(source).glob("*.py"):
        if module_path.name in ["__init__.py", "base.py"]:
            continue

        print(f"Processing {module_path}...")

        with Path.open(module_path) as fp:
            module_dump = fp.read()

        # Adjust imports.
        for relative_import in [".base", "..client", "..knowledge", "..model"]:
            module_dump = module_dump.replace(f"from {relative_import}", f"from .{relative_import}")

        # Modify function definitions.
        module_dump = re.sub(r"( {4}def )(?!_)", r"    async def ", module_dump)

        # Modify function calls.
        module_dump = re.sub(r"self\.client\.(.+)\(", r"await self.client.\1(", module_dump)
        module_dump = re.sub(r"= self\.(.+)\(", r"= await self.\1(", module_dump)
        module_dump = re.sub(r"send_request\(", r"await send_request(", module_dump)

        # Modify property accesses.
        module_dump = module_dump.replace("self.api.version", "await self.api.version")

        module_processed.append(module_path)
        target_path = Path(str(module_path).replace(str(source), str(target)))
        with Path.open(target_path, "w") as fp:
            fp.write(module_dump)

    relevant_modules = [_.name for _ in module_processed]
    existing_modules = [_.name for _ in Path(target).glob("*.py") if _.name not in ["base.py", "__init__.py"]]

    remove_module_count = 0

    for existing_module in existing_modules:
        if existing_module not in relevant_modules:
            print(f"Removing module {existing_module}...")
            (Path(target) / existing_module).unlink()
            remove_module_count += 1

    if not remove_module_count:
        print("No modules to remove.. pursuing..")

    with Path.open(source / "__init__.py") as fp:
        top_level_content = fp.read()

    print("Updating _async top level import content")

    top_level_content_patch = []

    for line in top_level_content.splitlines():
        if line.startswith("from "):
            if ".base" in line:
                continue

            patch_import_line = ""

            line_split = line.split(" ")

            for idx, word in zip(range(0, len(line_split)), line_split):
                if idx <= 2:
                    patch_import_line += word
                else:
                    patch_import_line += f"{word.replace(',', '')} as Async{word}"

                if idx != len(line_split) - 1:
                    patch_import_line += " "

            line = patch_import_line

        elif line.startswith(" " * 4):
            if "Base" in line:
                continue
            line = line.replace('    "', '    "Async')

        top_level_content_patch.append(line)

    with Path.open(target / "__init__.py", "w") as fp:
        fp.write("\n".join(top_level_content_patch) + "\n")

    # Run Ruff for code formatting, providing the same configuration as the project.
    shutil.copy(PYPROJECT_TOML, f"{target}")
    subprocess.call(["ruff", "format", target])
    subprocess.call(["ruff", "check", "--fix", target])
    Path(f"{target}/pyproject.toml").unlink()


def msg(text: str):
    """
    Return a colorful message, the color is determined by its text.
    """
    green = "\033[92;1m"
    red = "\033[91;1m"
    reset = "\033[0m"
    color = ""
    if text.lower().startswith("info"):
        color = green
    elif text.lower().startswith("error"):
        color = red
    return f"{color}{text}{reset}"


if __name__ == "__main__":
    subcommand = sys.argv[1:] and sys.argv[1] or None
    run(subcommand)



import unittest

from grafana_client.model import DatasourceHealthResponse

dhr = DatasourceHealthResponse(
    uid="uid",
    type="type",
    success=True,
    status="status",
    message="message",
    duration=42.42,
    response={},
)


class ModelTestCase(unittest.TestCase):
    def test_datasource_health_response_asdict(self):
        data = dhr.asdict()
        self.assertIsInstance(data, dict)
        self.assertTrue(len(data) > 1)
        self.assertIn("response", data)

    def test_datasource_health_response_asdict_compact(self):
        data = dhr.asdict_compact()
        self.assertIsInstance(data, dict)
        self.assertTrue(len(data) > 1)
        self.assertNotIn("response", data)



from ..base import Base


class Admin(Base):
    def __init__(self, client):
        super(Admin, self).__init__(client)
        self.client = client

    async def settings(self):
        """

        :return:
        """
        path = "/admin/settings"
        return await self.client.GET(path)

    async def stats(self):
        """

        :return:
        """
        path = "/admin/stats"
        return await self.client.GET(path)

    async def create_user(self, user):
        """

        :param user:
        :return:
        """
        create_user_path = "/admin/users"
        return await self.client.POST(create_user_path, json=user)

    async def change_user_password(self, user_id, password):
        """

        :param user_id:
        :param password:
        :return:
        """
        change_user_password_path = "/admin/users/%s/password" % user_id
        return await self.client.PUT(change_user_password_path, json={"password": password})

    async def change_user_permissions(self, user_id, is_grafana_admin):
        """

        :param user_id:
        :param is_grafana_admin:
        :return:
        """
        change_user_permissions = "/admin/users/%s/permissions" % user_id
        return await self.client.PUT(change_user_permissions, json={"isGrafanaAdmin": is_grafana_admin})

    async def delete_user(self, user_id):
        """

        :param user_id:
        :return:
        """
        delete_user_path = "/admin/users/%s" % user_id
        return await self.client.DELETE(delete_user_path)

    async def pause_all_alerts(self, pause):
        """

        :param pause:
        :return:
        """
        change_user_permissions = "/admin/pause-all-alerts"
        return await self.client.POST(change_user_permissions, json={"paused": pause})

    async def set_user_enabled(self, user_id, enabled: bool):
        """

        :param user_id:
        :param enabled:
        :return:
        """
        action = "enable" if enabled else "disable"
        set_user_enabled = "/admin/users/%s/%s" % (user_id, action)
        return await self.client.POST(set_user_enabled)



import unittest

from grafana_client import GrafanaApi

from ..compat import requests_mock


class ServiceAccountsTestCase(unittest.TestCase):
    def setUp(self):
        self.grafana = GrafanaApi(("admin", "admin"), host="localhost", url_path_prefix="", protocol="http")

    @requests_mock.Mocker()
    def test_get(self, m):
        m.get(
            "http://localhost/api/serviceaccounts/42?accesscontrol=true",
            json={
                "id": 42,
                "name": "rjuan",
                "login": "sa-juan",
                "orgId": 1,
                "isDisabled": False,
                "createdAt": "2023-12-15T21:38:59Z",
                "updatedAt": "2023-12-19T12:46:06Z",
                "avatarUrl": "/avatar/06b4d9db72de4d293813135da09fd736",
                "role": "Viewer",
                "teams": None,
                "accessControl": {
                    "serviceaccounts.permissions:read": True,
                    "serviceaccounts.permissions:write": True,
                    "serviceaccounts:delete": True,
                    "serviceaccounts:read": True,
                    "serviceaccounts:write": True,
                },
            },
        )
        result = self.grafana.serviceaccount.get(42)
        self.assertEqual(
            result,
            {
                "id": 42,
                "name": "rjuan",
                "login": "sa-juan",
                "orgId": 1,
                "isDisabled": False,
                "createdAt": "2023-12-15T21:38:59Z",
                "updatedAt": "2023-12-19T12:46:06Z",
                "avatarUrl": "/avatar/06b4d9db72de4d293813135da09fd736",
                "role": "Viewer",
                "teams": None,
                "accessControl": {
                    "serviceaccounts.permissions:read": True,
                    "serviceaccounts.permissions:write": True,
                    "serviceaccounts:delete": True,
                    "serviceaccounts:read": True,
                    "serviceaccounts:write": True,
                },
            },
        )

    @requests_mock.Mocker()
    def test_create(self, m):
        m.post(
            "http://localhost/api/serviceaccounts/",
            json={"message": "Service account created"},
        )
        user = self.grafana.serviceaccount.create({"name": "foo", "role": "Admin"})
        self.assertEqual(user["message"], "Service account created")

    @requests_mock.Mocker()
    def test_update(self, m):
        m.patch(
            "http://localhost/api/serviceaccounts/42",
            json={"message": "Service account updated"},
        )
        user = self.grafana.serviceaccount.update(42, {"name": "foo", "role": "Admin"})
        self.assertEqual(user["message"], "Service account updated")

    @requests_mock.Mocker()
    def test_delete(self, m):
        m.delete(
            "http://localhost/api/serviceaccounts/42",
            json={"message": "Service account deleted"},
        )
        user = self.grafana.serviceaccount.delete(42)
        self.assertEqual(user["message"], "Service account deleted")

    @requests_mock.Mocker()
    def test_create_token(self, m):
        m.post(
            "http://localhost/api/serviceaccounts/42/tokens",
            json={"message": "Service account token created"},
        )
        user = self.grafana.serviceaccount.create_token(42, {"name": "some-uuid"})
        self.assertEqual(user["message"], "Service account token created")

    @requests_mock.Mocker()
    def test_delete_token(self, m):
        m.delete(
            "http://localhost/api/serviceaccounts/42/tokens/2",
            json={"message": "Service account token deleted"},
        )
        user = self.grafana.serviceaccount.delete_token(42, 2)
        self.assertEqual(user["message"], "Service account token deleted")

    @requests_mock.Mocker()
    def test_get_tokens_some(self, m):
        m.get(
            "http://localhost/api/serviceaccounts/42/tokens",
            json=["token1", "token2"],
        )
        result = self.grafana.serviceaccount.get_tokens(42)
        self.assertEqual(len(result), 2)

    @requests_mock.Mocker()
    def test_get_tokens_zero(self, m):
        m.get(
            "http://localhost/api/serviceaccounts/42/tokens",
            json=[],
        )
        result = self.grafana.serviceaccount.get_tokens(42)
        self.assertEqual(len(result), 0)

    @requests_mock.Mocker()
    def test_search(self, m):
        # TODO: Don't know how the shape of the response looks like.
        m.get(
            "http://localhost/api/serviceaccounts/search?query=foo&page=3&perpage=10",
            json={"foo": "bar"},
        )
        result = self.grafana.serviceaccount.search("foo", page=3, perpage=10)
        self.assertEqual(result, [{"foo": "bar"}])

    @requests_mock.Mocker()
    def test_search_one_success(self, m):
        m.get(
            "http://localhost/api/serviceaccounts/search?query=foo&page=1",
            json={"totalCount": 1, "serviceAccounts": [{"foo": "bar"}]},
        )
        m.get(
            "http://localhost/api/serviceaccounts/search?query=foo&page=2",
            json={"totalCount": 1, "serviceAccounts": []},
        )
        result = self.grafana.serviceaccount.search_one("foo")
        self.assertEqual(result, {"foo": "bar"})

    @requests_mock.Mocker()
    def test_search_one_find_two(self, m):
        m.get(
            "http://localhost/api/serviceaccounts/search?query=foo&page=1",
            json={"totalCount": 2, "serviceAccounts": [{"foo": "bar"}, {"foo": "baz"}]},
        )
        m.get(
            "http://localhost/api/serviceaccounts/search?query=foo&page=2",
            json={"totalCount": 2, "serviceAccounts": []},
        )
        with self.assertRaises(ValueError) as ex:
            self.grafana.serviceaccount.search_one("foo")
        self.assertEqual("More than one service account matched", str(ex.exception))

    @requests_mock.Mocker()
    def test_search_one_find_zero(self, m):
        m.get(
            "http://localhost/api/serviceaccounts/search?query=foo&page=1",
            json={"totalCount": 0, "serviceAccounts": []},
        )
        with self.assertRaises(ValueError) as ex:
            self.grafana.serviceaccount.search_one("foo")
        self.assertEqual("No service account matched", str(ex.exception))



import json
import logging
import time
import warnings
from typing import Dict, Optional, Tuple, Union
from urllib.parse import urlencode

from niquests import ReadTimeout
from verlib2 import Version

from ...client import GrafanaBadInputError, GrafanaClientError, GrafanaServerError
from ...knowledge import get_healthcheck_expression, query_factory
from ...model import DatasourceHealthResponse, DatasourceIdentifier
from ..base import Base

logger = logging.getLogger(__name__)

VERSION_10_2_2 = Version("10.2.2")
VERSION_9 = Version("9")
VERSION_8 = Version("8")
VERSION_7 = Version("7")
VERBOSE = False


class Datasource(Base):
    def __init__(self, client, api):
        super(Datasource, self).__init__(client)
        self.client = client
        self.api = api

    async def health(self, datasource_uid: str):
        """
        Makes a call to the health endpoint of a data source identified by the
        given ``uid``.

        Available in Grafana 9+.

        https://grafana.com/docs/grafana/latest/developers/http_api/data_source/#check-data-source-health
        """
        path = f"/datasources/uid/{datasource_uid}/health"
        return await self.client.GET(path)

    async def find_datasource(self, datasource_name):
        """

        :param datasource_name:
        :return:
        """
        get_datasource_path = "/datasources/name/%s" % datasource_name
        return await self.client.GET(get_datasource_path)

    async def get_datasource_by_id(self, datasource_id):
        """
        Warning: This API is deprecated since Grafana v9.0.0 and will be
                 removed in a future release.

        :param datasource_id:
        :return:
        """
        get_datasource_path = "/datasources/%s" % datasource_id
        return await self.client.GET(get_datasource_path)

    async def get_datasource_by_name(self, datasource_name):
        """

        :param datasource_name:
        :return:
        """
        get_datasource_path = "/datasources/name/%s" % datasource_name
        return await self.client.GET(get_datasource_path)

    async def get_datasource_by_uid(self, datasource_uid):
        """

        :param datasource_uid:
        :return:
        """
        get_datasource_path = "/datasources/uid/%s" % datasource_uid
        return await self.client.GET(get_datasource_path)

    async def get_datasource_id_by_name(self, datasource_name):
        """

        :param datasource_name:
        :return:
        """
        get_datasource_path = "/datasources/id/%s" % datasource_name
        return await self.client.GET(get_datasource_path)

    async def get(self, dsident: DatasourceIdentifier):
        """
        Get dashboard by either datasource_id, datasource_uid, or datasource_name.
        """
        if dsident.id:
            datasource = await self.get_datasource_by_id(dsident.id)
        elif dsident.uid:
            datasource = await self.get_datasource_by_uid(dsident.uid)
        elif dsident.name:
            datasource = await self.get_datasource_by_name(dsident.name)
        else:
            raise KeyError("Data source must be identified by one of id, uid, or name")
        return datasource

    async def create_datasource(self, datasource):
        """

        :param datasource:
        :return:
        """
        create_datasources_path = "/datasources"
        return await self.client.POST(create_datasources_path, json=datasource)

    async def update_datasource(self, datasource_id, datasource):
        """

        :param datasource_id:
        :param datasource:
        :return:
        """
        update_datasource = "/datasources/%s" % datasource_id
        return await self.client.PUT(update_datasource, json=datasource)

    async def update_datasource_by_uid(self, datasource_uid, datasource):
        """

        :param datasource_uid:
        :param datasource:
        :return:
        """
        update_datasource = "/datasources/uid/%s" % datasource_uid
        return await self.client.PUT(update_datasource, json=datasource)

    async def list_datasources(self):
        """

        :return:
        """
        list_datasources_path = "/datasources"
        return await self.client.GET(list_datasources_path)

    async def delete_datasource_by_id(self, datasource_id):
        """
        Warning: This API is deprecated since Grafana v9.0.0 and will be
                 removed in a future release.

        :param datasource_id:
        :return:
        """
        delete_datasource = "/datasources/%s" % datasource_id
        return await self.client.DELETE(delete_datasource)

    async def delete_datasource_by_name(self, datasource_name):
        """

        :param datasource_name:
        :return:
        """
        delete_datasource = "/datasources/name/%s" % datasource_name
        return await self.client.DELETE(delete_datasource)

    async def delete_datasource_by_uid(self, datasource_uid):
        """

        :param datasource_uid:
        :return:
        """
        delete_datasource = "/datasources/uid/%s" % datasource_uid
        return await self.client.DELETE(delete_datasource)

    async def enable_datasource_permissions(self, datasource_id):
        """
        The Data Source Permissions is only available in Grafana Enterprise.

        :param datasource_id:
        :return:
        """
        if Version(await self.api.version) > VERSION_10_2_2:
            raise NotImplementedError("Deprecated since Grafana 10.2.3")

        get_datasource_path = "/datasources/%s/enable-permissions" % datasource_id
        return await self.client.POST(get_datasource_path)

    async def disable_datasource_permissions(self, datasource_id):
        """
        The Data Source Permissions is only available in Grafana Enterprise.

        :param datasource_id:
        :return:
        """
        if Version(await self.api.version) > VERSION_10_2_2:
            raise NotImplementedError("Deprecated since Grafana 10.2.3")

        get_datasource_path = "/datasources/%s/disable-permissions" % datasource_id
        return await self.client.POST(get_datasource_path)

    async def get_datasource_permissions(self, datasource_id):
        """
        The Data Source Permissions is only available in Grafana Enterprise.

        :param datasource_id:
        :return:
        """
        if Version(await self.api.version) > VERSION_10_2_2:
            raise NotImplementedError("Deprecated since Grafana 10.2.3, please use get_rbac_datasources()")

        get_datasource_path = "/datasources/%s/permissions" % datasource_id
        return await self.client.GET(get_datasource_path)

    async def add_datasource_permissions(self, datasource_id, permissions):
        """
        The Data Source Permissions is only available in Grafana Enterprise.

        :param datasource_id:
        :param permissions:
        :return:
        """
        if Version(await self.api.version) > VERSION_10_2_2:
            raise NotImplementedError("Deprecated since Grafana 10.2.3, please use set_rbac_datasources_*()")

        get_datasource_path = "/datasources/%s/permissions" % datasource_id
        return await self.client.POST(get_datasource_path, json=permissions)

    async def remove_datasource_permissions(self, datasource_id, permission_id):
        """
        The Data Source Permissions is only available in Grafana Enterprise.

        :param datasource_id:
        :param permission_id:
        :return:
        """
        if Version(await self.api.version) > VERSION_10_2_2:
            raise NotImplementedError("Deprecated since Grafana 10.2.3, please use set_rbac_datasources_*()")

        get_datasource_path = "/datasources/%s/permissions/%s" % (datasource_id, permission_id)
        return await self.client.DELETE(get_datasource_path)

    async def get_datasource_proxy_data(
        self,
        datasource_id,
        query_type="query",
        version="v1",  # noqa: ARG002
        expr=None,
        time=None,
        start=None,
        end=None,
        step=None,
    ):
        """

        :param datasource_id:
        :param version: api_version currently v1
        :param query_type: query_range |query
        :param expr: expr to query

        :return: r (dict)
        """
        warnings.warn(
            DeprecationWarning(
                "The function `get_datasource_proxy_data()` will be removed in a future release. "
                "Please use the functions `query()`, `query_range()`, or `series()` instead."
            )
        )
        if query_type == "query":
            return self.query(datasource_id=datasource_id, query=expr, timestamp=time)
        elif query_type == "query_range":
            return self.query_range(datasource_id=datasource_id, query=expr, start=start, end=end, step=step)
        else:
            raise KeyError(f"Unknown or invalid query type: {query_type}")

    async def query(self, datasource_id, query, timestamp, access="proxy"):
        """

        :param datasource_id:
        :param query:
        :param timestamp:
        :param access:
        :return:
        """
        post_query_path = "/datasources/%s/%s/api/v1/query" % (access, datasource_id)
        return await self.client.POST(
            post_query_path,
            data={
                "query": query,
                "time": timestamp,
            },
        )

    async def query_range(self, datasource_id, query, start, end, step, access="proxy"):
        """

        :param datasource_id:
        :param query:
        :param start:
        :param end:
        :param step:
        :param access:
        :return:
        """
        post_query_range_path = "/datasources/%s/%s/api/v1/query_range" % (access, datasource_id)
        return await self.client.POST(
            post_query_range_path, data={"query": query, "start": start, "end": end, "step": step}
        )

    async def series(self, datasource_id, match, start, end, access="proxy"):
        """

        :param datasource_id:
        :param match:
        :param start:
        :param end:
        :param access:
        :return:
        """
        post_series_path = "/datasources/%s/%s/api/v1/series" % (access, datasource_id)
        return await self.client.POST(post_series_path, data={"match[]": match, "start": start, "end": end})

    async def smartquery(
        self,
        datasource: Union[DatasourceIdentifier, Dict],
        expression: str,
        attrs: Optional[dict] = None,
        request: Optional[dict] = None,
    ):
        """
        Send a query to the designated data source and return its response.

        TODO: This is by far not complete. The `query_factory` function has to
            be made more elaborate in order to query different data source
                types.
        """

        if isinstance(datasource, DatasourceIdentifier):
            datasource = await self.get(datasource)

        datasource_id = datasource["id"]
        datasource_type = datasource["type"]
        datasource_dialect = datasource.get("jsonData", {}).get("version", "InfluxQL")
        access_type = datasource["access"]

        # Sanity checks.
        if not request and not expression:
            raise ValueError("request or expression must be given")
        elif not request:
            model = {
                "refId": "test",
            }
            if expression is not None and (attrs is None or (attrs is not None and "query" not in attrs)):
                model["query"] = expression
            if attrs is not None:
                model.update(attrs)
            request = query_factory(datasource, model)

        # Compute request method, body, and endpoint.
        if "method" in request and isinstance(request["method"], str):
            if request["method"] == "POST":
                send_request = self.client.POST
            else:
                send_request = self.client.GET

        logger.info(f"Submitting request: {request}")

        # Certain data sources like InfluxDB 1.x, still use the `/datasources/proxy` route.
        if datasource_type == "influxdb" and datasource_dialect == "InfluxQL":
            url = f"/datasources/proxy/{datasource_id}/query"
            url += "?" + urlencode(request["params"])
            request_kwargs = {"data": request["data"]}

        elif datasource_type == "graphite":
            url = f"/datasources/proxy/{datasource_id}/render"
            request_kwargs = {"data": request["data"]}

        # This case is very special. It is used for Elasticsearch and Testdata.
        elif "url" in request and model["url"].startswith("url://"):
            url = model["url"].replace("url://", "")
            url = url.format(
                datasource_id=datasource.get("id"),
                datasource_uid=datasource.get("uid"),
                database_name=datasource.get("database"),
            )
            request_kwargs = {}
            send_request = self.client.GET

        elif datasource_type in ("prometheus", "loki") and Version(await self.api.version) <= VERSION_7:
            if (
                "queries" in request["data"]
                and len(request["data"]["queries"]) > 0
                and "instant" in request["data"]["queries"][0]
                and request["data"]["queries"][0]["instant"]
            ):
                return self.query(
                    datasource.get("id"),
                    request["expr"],
                    request["data"]["to"],
                )
            else:
                return self.query_range(
                    datasource.get("id"),
                    request["expr"],
                    request["data"]["from"],
                    request["data"]["to"],
                    request["data"]["step"],
                )

        # For all others, use the generic data source communication endpoint.
        elif access_type in ["server", "proxy"]:
            url = "/ds/query"
            request_kwargs = {"json": request["data"]}

        else:
            raise NotImplementedError(f"Unable to submit query to data source with access type '{access_type}'")

        # Submit query.
        try:
            return await send_request(url, **request_kwargs)
        except (GrafanaClientError, GrafanaServerError) as ex:
            logger.error(
                f"Querying data source failed. id={datasource_id}, type={datasource_type}. "
                f"Reason: {ex}. Response: {ex.response or '<empty>'}"
            )
            raise

    async def health_check(self, datasource: Union[DatasourceIdentifier, Dict]) -> DatasourceHealthResponse:
        """
        Run a data source health check and return its success state, duration,
        and an (error) message.

        See `examples/datasource-health-check.py` for an example implementation.
        """

        if isinstance(datasource, DatasourceIdentifier):
            datasource = await self.get(datasource)

        datasource_uid = datasource["uid"]
        datasource_type = datasource["type"]
        datasource_dialect = datasource.get("jsonData", {}).get("version")
        access_type = datasource["access"]

        # Sanity checks.
        if access_type not in ["server", "proxy"]:
            raise NotImplementedError(f"Unable to inquire data source with access type '{access_type}'")

        # Resolve status query by database type.
        expression = get_healthcheck_expression(datasource_type, datasource_dialect)

        start = time.time()
        message = "Unknown error"
        try:
            response = await self.smartquery(datasource, expression)
            response_display = response
            if VERBOSE:  # pragma:nocover
                response_display = json.dumps(response, indent=2)
            logger.debug(f"Health check query response is: {response_display}")
            success = False

            # Elasticsearch has a special response format.
            if datasource_type == "elasticsearch":
                database_name = datasource["database"]
                if database_name in response:
                    try:
                        _ = response[database_name]["mappings"]["properties"]
                        message = "Success"
                        success = True
                    except KeyError as ex:
                        reason = f"{ex.__class__.__name__}: {ex}"
                        message = f"Invalid response. {reason}"
                else:
                    message = f"No response for database '{database_name}'"

            # When probed, those data sources only return their own representations?
            elif datasource_type in ["fetzerch-sunandmoon-datasource", "testdata"]:
                try:
                    _ = response["id"]
                    _ = response["uid"]
                    if datasource_type == "fetzerch-sunandmoon-datasource":
                        _ = response["jsonData"]["latitude"]
                        _ = response["jsonData"]["longitude"]
                    message = "Success"
                    success = True
                except KeyError as ex:
                    reason = f"{ex.__class__.__name__}: {ex}"
                    message = f"Invalid response. {reason}"

            # Graphite has a special response format.
            elif datasource_type == "graphite":
                try:
                    result = response[0]
                    _ = result["target"]
                    _ = result["datapoints"]
                    message = "Success"
                    success = True
                except (IndexError, KeyError) as ex:
                    reason = f"{ex.__class__.__name__}: {ex}"
                    message = f"Invalid response. {reason}"

            elif datasource_type == "loki":
                if await self.api.version and VERSION_7 <= Version(await self.api.version) < VERSION_8:
                    if "status" in response and response["status"] == "success":
                        message = "Success"
                        success = True
                    else:
                        message = response.get("message", "Unknown error")
                elif "results" in response and "test" in response["results"]:
                    message = "Success"
                    success = True
                elif "message" in response:
                    message = response["message"]

            # With OpenTSDB, a 200 OK response with empty body is just fine.
            elif datasource_type == "opentsdb":
                message = "Success"
                success = True

            # With Tempo, a 200 OK response with a non-empty body is probably just fine.
            elif datasource_type == "tempo":
                if len(response) >= 0:
                    message = "Success"
                    success = True

            # With Zipkin, a 200 OK response with a JSON body containing an empty array is probably just fine.
            elif datasource_type == "zipkin":
                if response == []:
                    message = "Success"
                    success = True

            # Generic case, where the response has a top-level `results` or `data` key.
            else:
                if "results" in response:
                    success, message = await self.parse_health_response_results(response=response)
                elif "data" in response:
                    success, message = await self.parse_health_response_data(response=response)
                else:
                    message = "Response lacks expected keys 'results' or 'data'"

        except (GrafanaBadInputError, GrafanaServerError, GrafanaClientError) as ex:
            success = False
            response = ex.response
            if isinstance(response, dict):
                if datasource_type == "elasticsearch":
                    error = response["error"]
                    status_code = response["status"]
                    if "root_cause" in error:
                        error = error["root_cause"][0]
                        error_type = error["type"]
                        error_reason = error["reason"]
                        message = f"Status: {status_code}. Type: {error_type}. Reason: {error_reason}"
                    else:
                        message = str(error)

                elif "results" in ex.response:
                    message = response["results"]["test"]["error"]
                else:
                    message = f"Unknown: {ex}. Response: {response}"
            else:
                message = str(ex)

        except ReadTimeout as ex:  # pragma:nocover
            message = str(ex)
            success = False
            response = None

        duration = round(time.time() - start, 4)
        if success:
            status = "OK"
        else:
            status = "ERROR"
            logger.warning(message)

        return DatasourceHealthResponse(
            uid=datasource_uid,
            type=datasource_type,
            success=success,
            status=status,
            message=message,
            duration=duration,
            response=response,
        )

    async def health_inquiry(self, datasource_uid: str) -> DatasourceHealthResponse:
        """
        Inquiry data source health. Try native method available since Grafana 9 first,
        and fall back to client-side implementation afterwards.
        """

        # Check if data source actually exists.
        try:
            datasource = await self.get(DatasourceIdentifier(uid=datasource_uid))
            datasource_type = datasource["type"]
            logger.debug(f"Data source information: {datasource}")
        except GrafanaClientError as ex:
            logger.error(f"Data source with UID '{datasource_uid}' does not exist: {ex}. Response: {ex.response}")
            if ex.status_code == 404:
                message = ex.response.get("message")
                return DatasourceHealthResponse(
                    uid=datasource_uid,
                    type=None,
                    success=False,
                    status="ERROR",
                    message=message,
                    duration=None,
                    response=ex.response,
                )
            else:
                raise

        # Check data source health.
        health = None

        # Use native Grafana 9+ data source health check.
        start = time.time()
        raised = True
        noop = False
        if await self.api.version and Version(await self.api.version) >= VERSION_9:
            try:
                health_native = await self.health(datasource_uid=datasource_uid)
                logger.debug(f"Response from native data source health check: {health_native}")
                status = health_native["status"]
                message = health_native["message"]
                success = status == "OK"
                response = health_native
                raised = False
            except (GrafanaClientError, GrafanaServerError, GrafanaBadInputError) as ex:
                logger.warning(
                    f"Native data source health check failed. uid={datasource_uid}, ex={ex}. "
                    f"Status: {ex.status_code}. Response: {ex.response}"
                )
                message = None
                if ex.status_code in [400]:
                    status = ex.response["status"]
                    message = ex.response["message"]
                    success = False
                    response = ex.response
                    raised = False

                # When Grafana 9+ server-side health checks are not implemented yet, Grafana mostly croaks with either
                # `404 Not Found`, `Server Error 503: Plugin unavailable`, or `Server Error 504: There was an issue
                # communicating with your instance`. Let's make this a noop in order to fall back to the client-side
                # implementation.
                elif ex.status_code in [404, 503]:  # 504
                    noop = True
                elif ex.status_code >= 500:
                    status = "FATAL"
                    message = f"{ex.__class__.__name__}: {ex}"
                    success = False
                    response = ex.response
                    raised = False

                else:
                    raise

                if "code" in ex.response:
                    message = f"[{ex.response['code']}] {message}"

            finally:
                if not noop:
                    if raised:
                        raise
                    else:
                        duration = round(time.time() - start, 4)
                        health = DatasourceHealthResponse(
                            uid=datasource_uid,
                            type=datasource_type,
                            success=success,
                            status=status,
                            message=message,
                            duration=duration,
                            response=response,
                        )

        if health is None:
            # Resolve data source by UID.
            datasource = await self.get(DatasourceIdentifier(uid=datasource_uid))
            # Run client-side health check.
            health = await self.health_check(datasource=datasource)

        return health

    @staticmethod
    async def parse_health_response_results(response: Dict) -> Tuple[bool, str]:
        success = False
        results = response["results"]
        if isinstance(results, dict):
            try:
                # The `refId` currently used is always `test`, see `knowledge.py`.
                # TODO: Change to `gcX`.
                result = results["test"]

                # Handle response in new DataFrame format.
                # Data frames are available in Grafana 7.0+, and replaced the Time series and Table structures
                # with a more generic data structure that can support a wider range of data types.
                # -- https://grafana.com/docs/grafana/latest/developers/plugins/data-frames/
                if "frames" in result:
                    if not isinstance(result["frames"], list):
                        raise TypeError("DataFrame response detected, but 'frames' is not a list")
                    try:
                        message = result["frames"][0]["schema"]["meta"]["executedQueryString"]
                    except (IndexError, KeyError):
                        message = "Success"
                    success = True

                # Handle response in previous format, where the `refId` is reflected on the top-level.
                elif "refId" in result:
                    try:
                        message = result["meta"]["executedQueryString"]
                    except KeyError:
                        message = "Success"
                    success = True
                else:
                    raise TypeError("Invalid response format")
            except TypeError as ex:
                reason = f"{ex.__class__.__name__}: {ex}"
                message = f"FATAL: Unable to decode result from dictionary-type response. {reason}"

        # Evaluate first item when `results` is a list.
        elif isinstance(results, list):
            try:
                result = results[0]
                if "error" in result:
                    message = result["error"]
                else:
                    _ = result["statement_id"]
                    _ = result["series"]
                    message = "Success"
                    success = True
            except (IndexError, KeyError) as ex:
                reason = f"{ex.__class__.__name__}: {ex}"
                message = f"FATAL: Unable to decode result from list-type response. {reason}"
        else:
            message = f"FATAL: Unknown response type '{type(results)}'. Expected: dictionary or list."

        return success, message

    @staticmethod
    async def parse_health_response_data(response: Dict) -> Tuple[bool, str]:
        """
        Response from Jaeger::

            {"data":["jaeger-query"],"total":1,"limit":0,"offset":0,"errors":null}

        Response from Loki::

            {"status":"success","data":["__name__"]}

        """
        success = False
        message = str(response["data"])
        if "errors" in response and response["errors"]:
            message = str(response["errors"])
        else:
            success = True

        return success, message



import logging

from grafana_client.client import GrafanaClientError

from ..base import Base


class Plugin(Base):
    def __init__(self, client):
        super(Plugin, self).__init__(client)
        self.client = client
        self.logger = logging.getLogger(__name__)

    async def list(self):
        """
        Return list of all installed plugins.
        """
        path = "/plugins?embedded=0"
        return await self.client.GET(path)

    async def by_id(self, plugin_id):
        """
        Return single plugin item selected by plugin identifier.
        """
        plugins = await self.list()
        return get_plugin_by_id(plugin_list=plugins, plugin_id=plugin_id)

    async def install(self, plugin_id, version=None, errors="raise"):
        """
        Install a 3rd-party plugin from the plugin store.
        """
        try:
            path = "/plugins/%s/install" % plugin_id
            # Unfortunately, this endpoint may respond with an empty JSON,
            # which needs compensation, because it does not decode well.
            return await self.client.POST(path, json={"version": version}, accept_empty_json=True)
        except GrafanaClientError as ex:
            # Ignore `Client Error 409: Plugin already installed`.
            if "409" not in str(ex):
                raise
        except Exception as ex:
            if errors == "raise":
                raise
            elif errors == "ignore":
                self.logger.warning(f"Problem installing plugin {plugin_id}: {ex}")
            else:
                raise ValueError(f"error={errors} is invalid")
        return None

    async def uninstall(self, plugin_id, errors="raise"):
        """
        Uninstall a 3rd-party plugin from the Grafana instance.
        """
        try:
            path = "/plugins/%s/uninstall" % plugin_id
            # Unfortunately, this endpoint may respond with an empty JSON,
            # which needs compensation, because it does not decode well.
            return await self.client.POST(path, accept_empty_json=True)
        except Exception as ex:
            if errors == "raise":
                raise
            elif errors == "ignore":
                self.logger.warning(f"Problem uninstalling plugin {plugin_id}: {ex}")
            else:
                raise ValueError(f"error={errors} is invalid")
        return None

    async def health(self, plugin_id):
        """
        Run a health check probe on the designated plugin.
        """
        path = "/plugins/%s/health" % plugin_id
        return await self.client.GET(path)

    async def metrics(self, plugin_id):
        """
        Inquire metrics of the designated plugin.
        """
        path = "/plugins/%s/metrics" % plugin_id
        return await self.client.GET(path)


def get_plugin_by_id(plugin_list, plugin_id):
    """
    Helper function to filter plugin list by identifier.
    """
    try:
        return next(item for item in plugin_list if item["id"] == plugin_id)
    except StopIteration:
        raise KeyError(f"Plugin not found: {plugin_id}")



try:
    from importlib.metadata import PackageNotFoundError, version
except ModuleNotFoundError:  # pragma:nocover
    from importlib_metadata import PackageNotFoundError, version

from .api import AsyncGrafanaApi, GrafanaApi  # noqa:E402,F401
from .client import HeaderAuth, TokenAuth  # noqa:E402,F401

__appname__ = "grafana-client"

try:
    __version__ = version(__appname__)
except PackageNotFoundError:  # pragma: no cover
    __version__ = "unknown"



import dataclasses
from typing import Any, Dict, Optional, Union


@dataclasses.dataclass
class DatasourceIdentifier:
    """
    A Grafana data source can be identified by either a numerical id, an
    alphanumerical uid, or a name.
    """

    id: Optional[str] = None
    uid: Optional[str] = None
    name: Optional[str] = None


@dataclasses.dataclass
class DatasourceModel:
    """
    Represent the minimum required fields to create a JSON payload suitable for
    submitting to the Grafana HTTP API.

    TODO: Field `database` will be deprecated.
          https://github.com/grafana/grafana/issues/59115
    """

    name: str
    type: str
    url: str
    access: str
    database: Optional[str] = None
    user: Optional[str] = None
    jsonData: Optional[Dict] = None
    secureJsonData: Optional[Dict] = None
    secureJsonFields: Optional[Dict] = None

    def asdict(self):
        return dataclasses.asdict(self)


@dataclasses.dataclass
class DatasourceHealthResponse:
    """
    Represent a response for a Grafana data source health check.

    It is a mixture of the fields of the native, server-side health check
    introduced with Grafana 9+, and the client-side implementation by
    `grafana-client`.

    - Grafana 9+ uses the fields `status` and `message`, where `status` can be
      one of `OK`, or `ERROR`.
    - `grafana-client` employs a more verbose response, including metadata
      about the requested data source item (`uid`, `type`), a boolean `success`
      flag, as well as the duration in seconds the request needed and the full
      response object.
    """

    uid: str
    type: Union[str, None]
    success: bool
    status: str
    message: str
    duration: Optional[float] = None
    response: Optional[Any] = None

    def asdict(self):
        return dataclasses.asdict(self)

    def asdict_compact(self):
        data = self.asdict()
        del data["response"]
        return data


@dataclasses.dataclass
class PersonalPreferences:
    """
    Request/response model for user-, team- and organization-preferences.

    https://grafana.com/docs/grafana/latest/developers/http_api/preferences/
    """

    homeDashboardId: Optional[int] = None
    homeDashboardUID: Optional[str] = None
    locale: Optional[str] = None
    theme: Optional[str] = None
    timezone: Optional[str] = None
    weekStart: Optional[str] = None

    def asdict(self, filter_none=False):
        if filter_none:
            return dataclasses.asdict(self, dict_factory=self.dict_factory_filter_none)
        else:
            return dataclasses.asdict(self)

    @staticmethod
    def dict_factory_filter_none(seq=None, **kwargs):
        seq = [item for item in seq if item[1] is not None]
        data = dict(seq)
        data.update(kwargs)
        return data



from verlib2 import Version

from ..base import Base

VERSION_8_2 = Version("8.2")


class LibraryElement(Base):
    Panel: int = 1
    Variable: int = 2

    def __init__(self, client, api):
        super(LibraryElement, self).__init__(client)
        self.client = client
        self.api = api

    async def get_library_element(self, element_uid: str) -> any:
        """

        :param element_uid:
        :return:
        """
        if await self.api.version and Version(await self.api.version) < VERSION_8_2:
            raise DeprecationWarning("Grafana versions earlier than 8.2 do not support library elements")
        get_element_path = f"/library-elements/{element_uid}"
        return await self.client.GET(get_element_path)

    async def get_library_element_by_name(self, element_name: str) -> any:
        """

        :param element_name:
        :return:
        """
        if await self.api.version and Version(await self.api.version) < VERSION_8_2:
            raise DeprecationWarning("Grafana versions earlier than 8.2 do not support library elements")
        get_element_path = f"/library-elements/name/{element_name}"
        return await self.client.GET(get_element_path)

    async def get_library_element_connections(self, element_uid: str) -> any:
        """

        :param element_uid:
        :return:
        """
        if await self.api.version and Version(await self.api.version) < VERSION_8_2:
            raise DeprecationWarning("Grafana versions earlier than 8.2 do not support library elements")
        get_element_connections_path = f"/library-elements/{element_uid}/connections"
        return await self.client.GET(get_element_connections_path)

    async def create_library_element(
        self, model: dict, name: str = None, kind: int = Panel, uid: str = None, folder_uid: str = None
    ):
        """

        :param model:
        :param name:
        :param kind:
        :param uid:
        :param folder_uid:
        :return:
        """
        if await self.api.version and Version(await self.api.version) < VERSION_8_2:
            raise DeprecationWarning("Grafana versions earlier than 8.2 do not support library elements")
        json: dict = dict()
        # If the model contains a "meta" entry, use the "folderUid" entry if folder_uid isn't given
        if folder_uid is not None:
            json["folderUid"] = folder_uid
        elif "meta" in model and "folderUid" in model["meta"]:
            json["folderUid"] = model["meta"]["folderUid"]
        # If the model contains a "model" entry, use the "uid" entry if uid isn't given
        if uid is not None:
            json["uid"] = uid
        elif "model" in model and "uid" in model["model"]:
            json["uid"] = model["model"]["uid"]
        # If the model contains a "model" entry, use the "title" entry if title isn't given
        if name is not None:
            json["name"] = name
        elif "model" in model and "title" in model["model"]:
            json["name"] = model["model"]["title"]
        if "model" in model:
            json["model"] = model["model"]
        else:
            json["model"] = model
        # If the model contains a "kind" entry, use that instead of the specified kind
        if "kind" in model:
            json["kind"] = model["kind"]
        else:
            json["kind"] = kind
        create_element_path = "/library-elements"
        return await self.client.POST(create_element_path, json=json)

    async def update_library_element(
        self, uid: str, model: dict, name: str = None, kind: int = Panel, folder_uid: str = None, version: int = None
    ):
        """

        :param uid:
        :param model:
        :param name:
        :param kind:
        :param folder_uid:
        :return:
        """
        if await self.api.version and Version(await self.api.version) < VERSION_8_2:
            raise DeprecationWarning("Grafana versions earlier than 8.2 do not support library elements")
        json: dict = dict()
        # If the model contains a "meta" entry, use the "folderUid" entry if folder_uid isn't given
        if folder_uid is not None:
            json["folderUid"] = folder_uid
        elif "meta" in model and "folderUid" in model["meta"]:
            json["folderUid"] = model["meta"]["folderUid"]
        json["uid"] = uid
        # If the model contains a "model" entry, use the "title" entry if title isn't given
        if name is not None:
            json["name"] = name
        elif "model" in model and "title" in model["model"]:
            json["name"] = model["model"]["title"]
        if "model" in model:
            json["model"] = model["model"]
        else:
            json["model"] = model
        # If the model contains a "kind" entry, use that instead of the specified kind
        if "kind" in model:
            json["kind"] = model["kind"]
        else:
            json["kind"] = kind
        if version is None:
            current_element = await self.get_library_element(uid)
            if "version" in current_element:
                json["version"] = current_element["version"]
            else:
                raise ValueError("version must be specified")
        else:
            json["version"] = version

        update_element_path = f"/library-elements/{uid}"
        return await self.client.PATCH(update_element_path, json=json)

    async def delete_library_element(self, element_uid: str) -> any:
        """

        :param element_uid:
        :return:
        """
        if await self.api.version and Version(await self.api.version) < VERSION_8_2:
            raise DeprecationWarning("Grafana versions earlier than 8.2 do not support library elements")
        delete_element_path = f"/library-elements/{element_uid}"
        return await self.client.DELETE(delete_element_path)

    async def list_library_elements(
        self,
        search_string: str = None,
        kind: int = None,
        sort_direction: str = None,
        type_filter: str = None,
        exclude_uid: str = None,
        folder_filter: str = None,
        per_page: int = None,
        page: int = None,
    ) -> any:
        """

        :param search_string:
        :param kind:
        :param sort_direction:
        :param type_filter:
        :param exclude_uid:
        :param folder_filter:
        :param per_page:
        :param page:

        :return:
        """
        if await self.api.version and Version(await self.api.version) < VERSION_8_2:
            raise DeprecationWarning("Grafana versions earlier than 8.2 do not support library elements")
        list_elements_path = "/library-elements"
        params = []

        if search_string is not None:
            params.append("searchString=%s" % search_string)
        if kind is not None:
            params.append("kind=%d" % kind)
        if sort_direction is not None:
            params.append("sortDirection=%s" % sort_direction)
        if type_filter is not None:
            params.append("typeFilter=%s" % type_filter)
        if exclude_uid is not None:
            params.append("excludeUid=%s" % exclude_uid)
        if folder_filter is not None:
            params.append("folderFilter=%s" % folder_filter)
        if per_page is not None:
            params.append("perPage=%d" % per_page)
        if page is not None:
            params.append("page=%d" % page)

        list_elements_path += "?"
        list_elements_path += "&".join(params)
        return await self.client.GET(list_elements_path)



from .admin import Admin as AsyncAdmin
from .alerting import Alerting as AsyncAlerting
from .alertingprovisioning import AlertingProvisioning as AsyncAlertingProvisioning
from .annotations import Annotations as AsyncAnnotations
from .dashboard import Dashboard as AsyncDashboard
from .dashboard_versions import DashboardVersions as AsyncDashboardVersions
from .datasource import Datasource as AsyncDatasource
from .folder import Folder as AsyncFolder
from .health import Health as AsyncHealth
from .libraryelement import LibraryElement as AsyncLibraryElement
from .notifications import Notifications as AsyncNotifications
from .organization import Organization as AsyncOrganization
from .organization import Organizations as AsyncOrganizations
from .plugin import Plugin as AsyncPlugin
from .rbac import Rbac as AsyncRbac
from .search import Search as AsyncSearch
from .service_account import ServiceAccount as AsyncServiceAccount
from .snapshots import Snapshots as AsyncSnapshots
from .team import Teams as AsyncTeams
from .user import User as AsyncUser
from .user import Users as AsyncUsers

__all__ = (
    "AsyncAdmin",
    "AsyncAlerting",
    "AsyncAlertingProvisioning",
    "AsyncAnnotations",
    "AsyncDashboard",
    "AsyncDashboardVersions",
    "AsyncDatasource",
    "AsyncFolder",
    "AsyncHealth",
    "AsyncLibraryElement",
    "AsyncNotifications",
    "AsyncOrganization",
    "AsyncOrganizations",
    "AsyncPlugin",
    "AsyncRbac",
    "AsyncSearch",
    "AsyncServiceAccount",
    "AsyncSnapshots",
    "AsyncTeams",
    "AsyncUser",
    "AsyncUsers",
)



import warnings

from verlib2 import Version

from .base import Base

VERSION_8 = Version("8")


class Dashboard(Base):
    def __init__(self, client, api):
        super(Dashboard, self).__init__(client)
        self.client = client
        self.api = api

    def get_dashboard(self, dashboard_uid):
        """

        :param dashboard_uid:
        :return:
        """
        get_dashboard_path = "/dashboards/uid/%s" % dashboard_uid
        return self.client.GET(get_dashboard_path)

    def get_dashboard_by_name(self, dashboard_name):
        """

        :param dashboard_name:
        :return:
        """
        if self.api.version and Version(self.api.version) >= VERSION_8:
            raise DeprecationWarning("Grafana 8 and higher does not support getting dashboards by slug")
        get_dashboard_path = "/dashboards/db/%s" % dashboard_name
        return self.client.GET(get_dashboard_path)

    def update_dashboard(self, dashboard):
        """

        :param dashboard:
        :return:
        """

        # When `folderId` or `folderUid` are not available within the dashboard payload,
        # populate them from the nested `meta` object, when given.
        for attribute in ["folderId", "folderUid"]:
            if attribute not in dashboard:
                if "meta" in dashboard and attribute in dashboard["meta"]:
                    dashboard = dashboard.copy()
                    dashboard[attribute] = dashboard["meta"][attribute]

        put_dashboard_path = "/dashboards/db"
        return self.client.POST(put_dashboard_path, json=dashboard)

    def delete_dashboard(self, dashboard_uid):
        """

        :param dashboard_uid:
        :return:
        """
        delete_dashboard_path = "/dashboards/uid/%s" % dashboard_uid
        return self.client.DELETE(delete_dashboard_path)

    def get_home_dashboard(self):
        """

        :return:
        """
        get_home_dashboard_path = "/dashboards/home"
        return self.client.GET(get_home_dashboard_path)

    def get_dashboards_tags(self):
        """

        :return:
        """
        get_dashboards_tags_path = "/dashboards/tags"
        return self.client.GET(get_dashboards_tags_path)

    def get_dashboard_permissions(self, dashboard_id):
        warnings.warn(
            "get_dashboard_permissions is deprecated, use corresponding _by_id or _by_uid methods",
            DeprecationWarning,
        )
        return self.get_permissions_by_id(dashboard_id)

    def update_dashboard_permissions(self, dashboard_id, items):
        warnings.warn(
            "update_dashboard_permissions is deprecated, use corresponding _by_id or _by_uid methods",
            DeprecationWarning,
        )
        return self.update_permissions_by_id(dashboard_id, items)

    def get_permissions_by_id(self, dashboard_id):
        return self.get_permissions_generic(dashboard_id, idtype="id")

    def update_permissions_by_id(self, dashboard_id, items):
        return self.update_permissions_generic(dashboard_id, items, idtype="id")

    def get_permissions_by_uid(self, dashboard_id):
        return self.get_permissions_generic(dashboard_id)

    def update_permissions_by_uid(self, dashboard_id, items):
        return self.update_permissions_generic(dashboard_id, items)

    def get_permissions_generic(self, identifier, idtype="uid"):
        permissions_path = f"/dashboards/{idtype}/{identifier}/permissions"
        return self.client.GET(permissions_path)

    def update_permissions_generic(self, identifier, items, idtype="uid"):
        permissions_path = f"/dashboards/{idtype}/{identifier}/permissions"
        return self.client.POST(permissions_path, json=items)
